# -*- coding: utf-8 -*-
"""
Unit tests for landlab.components.overland_flow.OverlandFlow

last updated: 3/14/16
"""
from nose.tools import assert_equal, assert_true, assert_raises, with_setup
try:
    from nose.tools import assert_is_instance
except ImportError:
    from landlab.testing.tools import assert_is_instance
import numpy as np

from landlab import RasterModelGrid
from landlab.components.overland_flow import OverlandFlow
from landlab.grid.structured_quad.links import left_edge_horizontal_ids

(_SHAPE, _SPACING, _ORIGIN) = ((32, 240), (25, 25), (0., 0.))
_ARGS = (_SHAPE, _SPACING, _ORIGIN)


def setup_grid():
        from landlab import RasterModelGrid
        grid = RasterModelGrid((32, 240), spacing = 25)
        grid.add_zeros('node', 'water__depth')
        grid.add_zeros('node', 'topographic__elevation')
        grid.add_zeros('water__discharge', at='active_link')
        deAlm = OverlandFlow(grid, mannings_n = 0.01, h_init=0.001)
        globals().update({
            'deAlm': OverlandFlow(grid)})


@with_setup(setup_grid)
def test_deAlm_name():
    assert_equal(deAlm.name, 'OverlandFlow')


@with_setup(setup_grid)
def test_deAlm_input_var_names():
    assert_equal(deAlm.input_var_names,  ('water__depth',
                                          'topographic__elevation',))


@with_setup(setup_grid)
def test_deAlm_output_var_names():
    assert_equal(deAlm.output_var_names, ('water__depth', 'water__discharge',
                                         'water_surface__gradient', ))

@with_setup(setup_grid)
def test_deAlm_var_units():
    assert_equal(set(deAlm.input_var_names) |
                 set(deAlm.output_var_names),
                 set(dict(deAlm.units).keys()))

    assert_equal(deAlm.var_units('water__depth'), 'm')
    assert_equal(deAlm.var_units('water__discharge'), 'm3/s')
    assert_equal(deAlm.var_units('water_surface__gradient'),'-')
    assert_equal(deAlm.var_units('topographic__elevation'), 'm')


@with_setup(setup_grid)
def test_grid_shape():
    assert_equal(deAlm.grid.number_of_node_rows, _SHAPE[0])
    assert_equal(deAlm.grid.number_of_node_columns, _SHAPE[1])


def test_deAlm_analytical():
    from landlab import RasterModelGrid
    grid = RasterModelGrid((32, 240), spacing = 25)
    grid.add_zeros('node', 'water__depth')
    grid.add_zeros('node', 'topographic__elevation')
    grid.set_closed_boundaries_at_grid_edges(True, True, True, True)
    left_inactive_ids = left_edge_horizontal_ids(grid.shape)
    deAlm = OverlandFlow(grid, mannings_n = 0.01, h_init=0.001)
    time = 0.0

    while time < 500:
        grid['link']['water__discharge'][left_inactive_ids] =   (grid['link'][
        'water__discharge'][left_inactive_ids + 1])
        dt = deAlm.calc_time_step()
        deAlm.overland_flow(dt)
        h_boundary = (((7./3.) * (0.01**2) * (0.4**3) *
                  time) ** (3./7.))
        grid.at_node['water__depth'][grid.nodes[1: -1, 1]] = h_boundary
        time += dt

    x = np.arange(0, ((grid.shape[1]) * grid.dx), grid.dx)
    h_analytical = (-(7./3.) * (0.01**2) * (0.4**2) * (x - (0.4 * 500)))

    h_analytical[np.where(h_analytical > 0)] = (h_analytical[np.where(
        h_analytical > 0)] ** (3./7.))
    h_analytical[np.where(h_analytical < 0)] = 0.0

    hdeAlm = deAlm.h.reshape(grid.shape)
    hdeAlm = hdeAlm[1][1:]
    hdeAlm = np.append(hdeAlm, [0])
    np.testing.assert_almost_equal(h_analytical, hdeAlm, decimal=1)


import numpy as np
from numpy.testing import assert_array_equal
from nose import with_setup
from nose.tools import assert_equal
try:
    from nose.tools import assert_is
except ImportError:
    from landlab.testing.tools import assert_is

from landlab.grid.raster_steepest_descent import (
    _calc_steepest_descent_across_cell_faces)


_GRID = None
_VALUES_AT_NODES = None


def setup_unit_grid():
    from landlab import RasterModelGrid
    global _GRID, _VALUES_AT_NODES
    _GRID = RasterModelGrid(4, 5)
    _VALUES_AT_NODES = np.arange(20.)


def setup_3x3_grid():
    from landlab import RasterModelGrid
    global rmg_3x3
    rmg_3x3 = RasterModelGrid(3, 3)


@with_setup(setup_unit_grid)
def test_scalar_arg():
    grad = _calc_steepest_descent_across_cell_faces(
        _GRID, _VALUES_AT_NODES, 0)
    assert_equal(grad, -5.)

    grad = _GRID._calc_steepest_descent_across_cell_faces(
        _VALUES_AT_NODES, 0)
    assert_equal(grad, -5.)


@with_setup(setup_unit_grid)
def test_iterable():
    grad = _calc_steepest_descent_across_cell_faces(
        _GRID, _VALUES_AT_NODES, [0, 4])
    assert_array_equal(grad, [-5., -5.])

    grad = _GRID._calc_steepest_descent_across_cell_faces(
        _VALUES_AT_NODES, [0, 4])
    assert_array_equal(grad, [-5., -5.])


@with_setup(setup_unit_grid)
def test_scalar_arg_with_return_node():
    values = np.array([0, 1,  3, 6, 10,
                       0, 1,  3, 6, 10,
                       0, 1,  3, 5, 10,
                       0, 1, -3, 6, 10, ], dtype=float)
    (grad, node) = _GRID._calc_steepest_descent_across_cell_faces(
        values, (0, 4), return_node=True)
    assert_array_equal(grad, [-1, -6])
    assert_array_equal(node, [5, 17])


@with_setup(setup_3x3_grid)
def test_node_in_direction_of_max():
    for neighbor_id in [1, 3, 5, 7]:
        values = np.zeros(9)
        values[neighbor_id] = -1
        (_, node) = rmg_3x3._calc_steepest_descent_across_cell_faces(
            values, 0, return_node=True)
        assert_array_equal(node, neighbor_id)


@with_setup(setup_3x3_grid)
def test_node_in_direction_of_max_with_ties():
    values = np.zeros(9)
    (_, node) = rmg_3x3._calc_steepest_descent_across_cell_faces(
        values, 0, return_node=True)
    assert_array_equal(node, 5)

    values = np.zeros(9)
    values[5] = 1
    (_, node) = rmg_3x3._calc_steepest_descent_across_cell_faces(
        values, 0, return_node=True)
    assert_array_equal(node, 7)

    values = np.zeros(9)
    values[[5, 7]] = 1
    (_, node) = rmg_3x3._calc_steepest_descent_across_cell_faces(
        values, 0, return_node=True)
    assert_array_equal(node, 3)



from landlab.field.scalar_data_fields import ScalarDataFields, FieldError
from landlab.field.grouped import ModelDataFields, GroupError, GroupSizeError
from landlab.field.field_mixin import ModelDataFieldsMixIn

__all__ = ['ScalarDataFields', 'ModelDataFields', 'ModelDataFieldsMixIn',
           'FieldError', 'GroupError', 'GroupSizeError']



# -*- coding: utf-8 -*-
"""
Unit tests for landlab.components.overland_flow.OverlandFlowBates

last updated: 3/14/16
"""
from nose.tools import assert_equal, assert_true, assert_raises, with_setup
try:
    from nose.tools import assert_is_instance
except ImportError:
    from landlab.testing.tools import assert_is_instance
import numpy as np

from landlab import RasterModelGrid
from landlab.components.overland_flow import OverlandFlowBates, OverlandFlow


(_SHAPE, _SPACING, _ORIGIN) = ((32, 240), (25, 25), (0., 0.))
_ARGS = (_SHAPE, _SPACING, _ORIGIN)


def setup_grid():
        from landlab import RasterModelGrid
        grid = RasterModelGrid((32, 240), spacing = 25)
        grid.add_zeros('node', 'water__depth')
        grid.add_zeros('node', 'topographic__elevation')
        bates = OverlandFlowBates(grid, mannings_n = 0.01, h_init=0.001)
        globals().update({
            'bates': bates})


@with_setup(setup_grid)
def test_Bates_name():
    assert_equal(bates.name, 'OverlandFlowBates')


@with_setup(setup_grid)
def test_Bates_input_var_names():
    # DEJH added sets to remove reliance on ordering
    assert_equal(set(bates.input_var_names),  set(('water__depth',
                                                   'topographic__elevation')))


@with_setup(setup_grid)
def test_Bates_output_var_names():
    # DEJH added sets to remove reliance on ordering
    assert_equal(set(bates.output_var_names), set(('water__depth',
                                                   'water__discharge',
                                                   'water_surface__gradient')))

@with_setup(setup_grid)
def test_Bates_var_units():
    assert_equal(set(bates.input_var_names) |
                 set(bates.output_var_names),
                 set(dict(bates.units).keys()))

    assert_equal(bates.var_units['water__depth'], 'm')
    assert_equal(bates.var_units['water__discharge'], 'm3/s')
    assert_equal(bates.var_units['water_surface__gradient'], 'm/m')
    assert_equal(bates.var_units['topographic__elevation'], 'm')


@with_setup(setup_grid)
def test_field_initialized_to_zero():
    for name in bates.grid['node'].keys():
        field = bates.grid['node'][name]
        if name != 'water__depth':
            assert_true(np.all(np.isclose(field, 0.)))
        else:
            assert_true(np.all(np.isclose(field, 0.001)))
            # this remains broken, and needs JA's attention


@with_setup(setup_grid)
def test_grid_shape():
    assert_equal(bates.grid.number_of_node_rows, _SHAPE[0])
    assert_equal(bates.grid.number_of_node_columns, _SHAPE[1])


def test_Bates_analytical():
    from landlab import RasterModelGrid
    grid = RasterModelGrid((32, 240), spacing = 25)
    grid.add_zeros('node', 'water__depth')
    grid.add_zeros('node', 'topographic__elevation')
    grid.set_closed_boundaries_at_grid_edges(True, True, True, True)
    bates = OverlandFlowBates(grid, mannings_n = 0.01, h_init=0.001)
    time = 0.0
    bates.dt = 1.0
    while time < 500:
        bates.overland_flow(grid)
        h_boundary = (((7./3.) * (0.01**2) * (0.4**3) *
                  time) ** (3./7.))
        grid.at_node['water__depth'][grid.nodes[1: -1, 1]] = h_boundary
        time += bates.dt

    x = np.arange(0, ((grid.shape[1]) * grid.dx), grid.dx)
    h_analytical = (-(7./3.) * (0.01**2) * (0.4**2) * (x - (0.4 * 500)))

    h_analytical[np.where(h_analytical > 0)] = (h_analytical[np.where(
        h_analytical > 0)] ** (3./7.))
    h_analytical[np.where(h_analytical < 0)] = 0.0

    hBates = bates.h.reshape(grid.shape)
    hBates = hBates[1][1:]
    hBates = np.append(hBates, [0])
    np.testing.assert_almost_equal(h_analytical, hBates, decimal=1)


import numpy as np
from numpy.testing import assert_array_equal
from nose import with_setup
try:
    from nose.tools import assert_is
except ImportError:
    from landlab.testing.tools import assert_is

from landlab import RasterModelGrid


_GRID = None
_VALUES_AT_NODES = None


def setup_unit_grid():
    """Set up a test grid with unit spacing."""
    global _GRID, _VALUES_AT_NODES
    _GRID = RasterModelGrid(4, 5)
    _VALUES_AT_NODES = np.arange(20.)


def setup_non_unit_grid():
    """Set up a test grid with non-unit spacing."""
    global _GRID, _VALUES_AT_NODES
    _GRID = RasterModelGrid(4, 5, 2.)
    _VALUES_AT_NODES = np.arange(20.)


def setup_non_square_grid():
    """Set up a test grid with different x and y spacing."""
    global _GRID, _VALUES_AT_NODES
    _GRID = RasterModelGrid((4, 5), spacing=(5, 2))
    _VALUES_AT_NODES = np.arange(20.)


@with_setup(setup_unit_grid)
def test_unit_spacing():
    """Test on a grid with unit spacing."""
    grads = _GRID.calc_grad_at_link(_VALUES_AT_NODES)
    assert_array_equal(
        grads,
        np.array([1, 1, 1, 1, 
                  5, 5, 5, 5, 5,
                  1, 1, 1, 1,
                  5, 5, 5, 5, 5,
                  1, 1, 1, 1,
                  5, 5, 5, 5, 5,
                  1, 1, 1, 1],
                 dtype=float))
    diffs = _GRID.calculate_diff_at_links(_VALUES_AT_NODES)
    assert_array_equal(grads, diffs)


@with_setup(setup_non_square_grid)
def test_non_unit_spacing():
    """Test on a grid with non-unit spacing."""
    grads = _GRID.calc_grad_at_link(_VALUES_AT_NODES)
    assert_array_equal(
        grads,
        np.array(
           [0.5, 0.5,  0.5,  0.5,
            1.0, 1.0,  1. ,  1. ,  1. ,  
            0.5,  0.5, 0.5,  0.5,
            1. ,  1. , 1. ,  1. ,  1. ,
            0.5,  0.5,  0.5,  0.5,
            1. ,  1. ,  1. ,  1. , 1. ,
            0.5,  0.5,  0.5,  0.5],
            dtype=float))
    diffs = _GRID.calc_diff_at_link(_VALUES_AT_NODES)
    assert_array_equal(diffs,
        np.array([1, 1, 1, 1,
                  5, 5, 5, 5, 5,
                  1, 1, 1, 1,
                  5, 5, 5, 5, 5,
                  1, 1, 1, 1,
                  5, 5, 5, 5, 5,
                  1, 1, 1, 1],
                 dtype=float))


@with_setup(setup_unit_grid)
def test_out_array():
    """Test using the out keyword."""
    grads = np.empty(31)
    rtn_grads = _GRID.calc_grad_at_link(_VALUES_AT_NODES, out=grads)
    assert_array_equal(
        grads,
        np.array([1, 1, 1, 1, 5, 5, 5, 5, 5, 1, 1, 1, 1, 5, 5,
                  5, 5, 5, 1, 1, 1, 1, 5, 5, 5, 5, 5, 1, 1, 1, 1],
                 dtype=float))
    assert_is(rtn_grads, grads)


@with_setup(setup_unit_grid)
def test_diff_out_array():
    """Test that return array is the out keyword."""
    diff = np.empty(31)
    rtn_diff = _GRID.calculate_diff_at_links(_VALUES_AT_NODES, out=diff)
    assert_array_equal(
        diff,
        np.array([1, 1, 1, 1, 5, 5, 5, 5, 5, 1, 1, 1, 1, 5, 5,
                  5, 5, 5, 1, 1, 1, 1, 5, 5, 5, 5, 5, 1, 1, 1, 1],
                 dtype=float))
    assert_is(rtn_diff, diff)



#! /usr/bin/env python
"""Read data from a GEBCO NetCDF file into a RasterModelGrid."""


try:
    import netCDF4 as nc4
except ImportError:
    import warnings
    warnings.warn('Unable to import netCDF4.', ImportWarning)

from scipy.io import netcdf as nc

from landlab import RasterModelGrid
from landlab.io.netcdf.errors import NotRasterGridError


_COORDINATE_NAMES = ['x_range', 'y_range', 'z_range', 'spacing', 'dimension']
_AXIS_NAMES = ['x', 'y']


def _read_netcdf_grid_shape(root):
    """Read the grid shape from a GEBCO NetCDF file.

    Parameters
    ----------
    root : netcdf_file
        A NetCDF file.

    Returns
    -------
    tuple of int
        The shape of the grid as number of rows, then columns.
    """
    return root.variables['dimension'][:]


def _read_netcdf_grid_spacing(root):
    """Read the grid spacing from a GEBCO NetCDF file.

    Parameters
    ----------
    root : netcdf_file
        A NetCDF file.

    Returns
    -------
    tuple of float
        The spacing of the grid between rows, then columns.
    """
    return root.variables['spacing'][:]


def _read_netcdf_structured_data(root):
    """Read the grid data from a GEBCO NetCDF file.

    Parameters
    ----------
    root : netcdf_file
        A NetCDF file.

    Returns
    -------
    dict
        The data fields as numpy arrays. Keys are the variable names, and
        values are the data.
    """
    fields = dict()
    for (name, var) in root.variables.items():
        if name not in _COORDINATE_NAMES:
            fields[name] = var[:] * var.scale_factor + var.add_offset
            fields[name].shape = (fields[name].size, )
    return fields


def read_netcdf(nc_file, just_grid=False):
    """Read a GEBCO-formatted NetCDF file.

    Reads the NetCDF file *nc_file*, and writes it to the fields of a new
    RasterModelGrid, which it then returns.  Check the names of the fields
    in the returned grid with grid.at_nodes.keys().

    Parameters
    ----------
    nc_file : str
        Name of the NetCDF file.
    just_grid : bool, optional
        If ``True``, just read the grid information and forget the data.
        Otherwise add the data as fields.

    Returns
    -------
    RasterModelGrid
        A newly-created :any:`RasterModelGrid`.
    """
    try:
        root = nc.netcdf_file(nc_file, 'r', version=2)
    except TypeError:
        root = nc4.Dataset(nc_file, 'r', format='NETCDF4')

    shape = _read_netcdf_grid_shape(root)
    spacing = _read_netcdf_grid_spacing(root)

    assert len(shape) == 2
    assert len(spacing) == 2
    if spacing[0] != spacing[1]:
        raise NotRasterGridError()

    grid = RasterModelGrid(shape, spacing=spacing)

    if not just_grid:
        fields = _read_netcdf_structured_data(root)
        for (name, values) in fields.items():
            grid.add_field('node', name, values)

    root.close()

    return grid



import numpy as np

from numpy.testing import assert_array_equal
from nose.tools import raises, assert_equal

from landlab.grid.structured_quad import nodes

from landlab.grid.base import CORE_NODE, FIXED_VALUE_BOUNDARY, CLOSED_BOUNDARY
from landlab.testing.tools import assert_array_is_int


def test_perimeter_nodes():
    node_ids = nodes.perimeter((4, 5))
    assert_array_equal(node_ids,
                       [0, 1, 2, 3, 4, 5, 9, 10, 14, 15, 16, 17, 18, 19])
    assert_array_is_int(node_ids)


def test_perimeter_status_default():
    node_status = nodes.status_with_perimeter_as_boundary((4, 5))
    F, C = CLOSED_BOUNDARY, CORE_NODE
    assert_array_equal(node_status,
                       [[F, F, F, F, F],
                        [F, C, C, C, F],
                        [F, C, C, C, F],
                        [F, F, F, F, F]])
    assert_array_is_int(node_status)


def test_perimeter_status_status_as_scalar():
    node_status = nodes.status_with_perimeter_as_boundary(
        (4, 5), node_status=CLOSED_BOUNDARY)
    B, C = CLOSED_BOUNDARY, CORE_NODE
    assert_array_equal(node_status,
                       [[B, B, B, B, B],
                        [B, C, C, C, B],
                        [B, C, C, C, B],
                        [B, B, B, B, B]])
    assert_array_is_int(node_status)


def test_perimeter_status_status_as_array():
    F, B, C = FIXED_VALUE_BOUNDARY, CLOSED_BOUNDARY, CORE_NODE
    status = [F, F, F, F, F, F, B, F, B, B, B, B, B, B]
    node_status = nodes.status_with_perimeter_as_boundary(
        (4, 5), node_status=status)
    assert_array_equal(node_status,
                       [[F, F, F, F, F],
                        [F, C, C, C, B],
                        [F, C, C, C, B],
                        [B, B, B, B, B]])
    assert_array_is_int(node_status)



#! /usr/bin/env python
"""Map values from one grid element to another.

Grid mapping functions
+++++++++++++++++++++++

.. autosummary::
    :toctree: generated/

    ~landlab.grid.mappers.map_link_head_node_to_link
    ~landlab.grid.mappers.map_link_tail_node_to_link
    ~landlab.grid.mappers.map_min_of_link_nodes_to_link
    ~landlab.grid.mappers.map_max_of_link_nodes_to_link
    ~landlab.grid.mappers.map_mean_of_link_nodes_to_link
    ~landlab.grid.mappers.map_value_at_min_node_to_link
    ~landlab.grid.mappers.map_value_at_max_node_to_link
    ~landlab.grid.mappers.map_node_to_cell
    ~landlab.grid.mappers.map_min_of_node_links_to_node
    ~landlab.grid.mappers.map_max_of_node_links_to_node
    ~landlab.grid.mappers.map_upwind_node_link_max_to_node
    ~landlab.grid.mappers.map_downwind_node_link_max_to_node
    ~landlab.grid.mappers.map_upwind_node_link_mean_to_node
    ~landlab.grid.mappers.map_downwind_node_link_mean_to_node
    ~landlab.grid.mappers.map_value_at_upwind_node_link_max_to_node
    ~landlab.grid.mappers.map_value_at_downwind_node_link_max_to_node
    ~landlab.grid.mappers.dummy_func_to_demonstrate_docstring_modification

Each link has a *tail* and *head* node. The *tail* nodes are located at the
start of a link, while the head nodes are located at end of a link.

Below, the numbering scheme for links in `RasterModelGrid` is illustrated
with an example of a four-row by five column grid (4x5). In this example,
each * (or X) is a node, the lines represent links, and the ^ and > symbols
indicate the direction and *head* of each link. Link heads in the
`RasterModelGrid` always point in the cardinal directions North (N) or East
(E).::

    *--27-->*--28-->*--29-->*--30-->*
    ^       ^       ^       ^       ^
   22      23      24      25      26
    |       |       |       |       |
    *--18-->*--19-->*--20-->*--21-->*
    ^       ^       ^       ^       ^
    13      14      15      16     17
    |       |       |       |       |
    *---9-->*--10-->X--11-->*--12-->*
    ^       ^       ^       ^       ^
    4       5       6       7       8
    |       |       |       |       |
    *--0--->*---1-->*--2--->*---3-->*

For example, node 'X' has four link-neighbors. From south and going clockwise,
these neighbors are [6, 10, 15, 11]. Both link 6 and link 10 have node 'X' as
their 'head' node, while links 15 and 11 have node 'X' as their tail node.
"""
from __future__ import division

import numpy as np


def map_link_head_node_to_link(grid, var_name, out=None):
    """Map values from a link head nodes to links.

    Iterate over a grid and identify the node at the *head*. For each link,
    the value of *var_name* at the *head* node is mapped to the corresponding
    link.

    In a RasterModelGrid, each one node has two adjacent "link heads". This
    means each node value is mapped to two corresponding links.

    Construction::

        map_link_head_node_to_link(grid, var_name, out=None)

    Parameters
    ----------
    grid : ModelGrid
        A landlab ModelGrid.
    var_name : array or field name
        Values defined at nodes.
    out : ndarray, optional
        Buffer to place mapped values into or `None` to create a new array.

    Returns
    -------
    ndarray
        Mapped values at links.

    Examples
    --------
    >>> import numpy as np
    >>> from landlab.grid.mappers import map_link_head_node_to_link
    >>> from landlab import RasterModelGrid

    >>> rmg = RasterModelGrid((3, 4))
    >>> rmg.at_node['z'] = np.array([ 0,  1,  2,  3,
    ...                               4,  5,  6,  7,
    ...                               8,  9, 10, 11])
    >>> map_link_head_node_to_link(rmg, 'z')
    array([  1.,   2.,   3.,   4.,   5.,   6.,   7.,   5.,   6.,   7.,   8.,
             9.,  10.,  11.,   9.,  10.,  11.])

    >>> values_at_links = rmg.empty(at='link')
    >>> rtn = map_link_head_node_to_link(rmg, 'z', out=values_at_links)
    >>> values_at_links
    array([  1.,   2.,   3.,   4.,   5.,   6.,   7.,   5.,   6.,   7.,   8.,
             9.,  10.,  11.,   9.,  10.,  11.])
    >>> rtn is values_at_links
    True
    """
    if type(var_name) is str:
        var_name = grid.at_node[var_name]
    if out is None:
        out = grid.empty(at='link')
    out[:] = var_name[grid.node_at_link_head]

    return out


def map_link_tail_node_to_link(grid, var_name, out=None):
    """Map values from a link tail nodes to links.

    map_link_tail_node_to_link iterates across the grid and
    identifies the node at the "tail", or the "from" node for each link. For
    each link, the value of 'var_name' at the "from" node is mapped to the
    corresponding link.

    In a RasterModelGrid, each one node has two adjacent "link tails". This
    means each node value is mapped to two corresponding links.

    Construction::

        map_link_tail_node_to_link(grid, var_name, out=None)

    Parameters
    ----------
    grid : ModelGrid
        A landlab ModelGrid.
    var_name : array or field name
        Values defined at nodes.
    out : ndarray, optional
        Buffer to place mapped values into or `None` to create a new array.

    Returns
    -------
    ndarray
        Mapped values at links.

    Examples
    --------
    >>> import numpy as np
    >>> from landlab.grid.mappers import map_link_tail_node_to_link
    >>> from landlab import RasterModelGrid

    >>> rmg = RasterModelGrid((3, 4))
    >>> rmg.at_node['z'] = np.array([ 0,  1,  2,  3,
    ...                               4,  5,  6,  7,
    ...                               8,  9, 10, 11])
    >>> map_link_tail_node_to_link(rmg, 'z')
    array([  0.,   1.,   2.,   0.,   1.,   2.,   3.,   4.,   5.,   6.,   4.,
             5.,   6.,   7.,   8.,   9.,  10.])

    >>> values_at_links = rmg.empty(at='link')
    >>> rtn = map_link_tail_node_to_link(rmg, 'z', out=values_at_links)
    >>> values_at_links
    array([  0.,   1.,   2.,   0.,   1.,   2.,   3.,   4.,   5.,   6.,   4.,
             5.,   6.,   7.,   8.,   9.,  10.])
    >>> rtn is values_at_links
    True
    """
    if out is None:
        out = grid.empty(at='link')

    if type(var_name) is str:
        var_name = grid.at_node[var_name]
    out[:] = var_name[grid.node_at_link_tail]

    return out


def map_min_of_link_nodes_to_link(grid, var_name, out=None):
    """Map the minimum of a link's nodes to the link.

    map_min_of_link_nodes_to_link iterates across the grid and
    identifies the node values at both the "head" and "tail" of a given link.
    This function evaluates the value of 'var_name' at both the "to" and
    "from" node. The minimum value of the two node values is then mapped to
    the link.

    Construction::

        map_min_of_link_nodes_to_link(grid, var_name, out=None)

    Parameters
    ----------
    grid : ModelGrid
        A landlab ModelGrid.
    var_name : array or field name
        Values defined at nodes.
    out : ndarray, optional
        Buffer to place mapped values into or `None` to create a new array.

    Returns
    -------
    ndarray
        Mapped values at links.

    Examples
    --------
    >>> import numpy as np
    >>> from landlab.grid.mappers import map_min_of_link_nodes_to_link
    >>> from landlab import RasterModelGrid

    >>> rmg = RasterModelGrid((3, 4))
    >>> _ = rmg.add_field('node', 'z',
    ...                   [[ 0,  1,  2,  3],
    ...                    [ 7,  6,  5,  4],
    ...                    [ 8,  9, 10, 11]])
    >>> map_min_of_link_nodes_to_link(rmg, 'z')
    array([  0.,   1.,   2.,   0.,   1.,   2.,   3.,   6.,   5.,   4.,   7.,
             6.,   5.,   4.,   8.,   9.,  10.])

    >>> values_at_links = rmg.empty(at='link')
    >>> rtn = map_min_of_link_nodes_to_link(rmg, 'z', out=values_at_links)
    >>> values_at_links
    array([  0.,   1.,   2.,   0.,   1.,   2.,   3.,   6.,   5.,   4.,   7.,
             6.,   5.,   4.,   8.,   9.,  10.])
    >>> rtn is values_at_links
    True
    """
    if out is None:
        out = grid.empty(at='link')

    if type(var_name) is str:
        var_name = grid.at_node[var_name]
    np.minimum(var_name[grid.node_at_link_head],
               var_name[grid.node_at_link_tail], out=out)

    return out


def map_max_of_link_nodes_to_link(grid, var_name, out=None):
    """Map the maximum of a link's nodes to the link.

    map_max_of_link_nodes_to_link iterates across the grid and
    identifies the node values at both the "head" and "tail" of a given link.
    This function evaluates the value of 'var_name' at both the "to" and
    "from" node. The maximum value of the two node values is then mapped to
    the link.

    Construction::

        map_max_of_link_nodes_to_link(grid, var_name, out=None)

    Parameters
    ----------
    grid : ModelGrid
        A landlab ModelGrid.
    var_name : array or field name
        Values defined at nodes.
    out : ndarray, optional
        Buffer to place mapped values into or `None` to create a new array.

    Returns
    -------
    ndarray
        Mapped values at links.

    Examples
    --------
    >>> import numpy as np
    >>> from landlab.grid.mappers import map_max_of_link_nodes_to_link
    >>> from landlab import RasterModelGrid

    >>> rmg = RasterModelGrid((3, 4))
    >>> _ = rmg.add_field('node', 'z',
    ...                   [[0, 1, 2, 3],
    ...                    [7, 6, 5, 4],
    ...                    [8, 9, 10, 11]])
    >>> map_max_of_link_nodes_to_link(rmg, 'z')
    array([  1.,   2.,   3.,   7.,   6.,   5.,   4.,   7.,   6.,   5.,   8.,
             9.,  10.,  11.,   9.,  10.,  11.])

    >>> values_at_links = rmg.empty(at='link')
    >>> rtn = map_max_of_link_nodes_to_link(rmg, 'z', out=values_at_links)
    >>> values_at_links
    array([  1.,   2.,   3.,   7.,   6.,   5.,   4.,   7.,   6.,   5.,   8.,
             9.,  10.,  11.,   9.,  10.,  11.])
    >>> rtn is values_at_links
    True
    """
    if out is None:
        out = grid.empty(at='link')

    if type(var_name) is str:
        var_name = grid.at_node[var_name]
    np.maximum(var_name[grid.node_at_link_head],
               var_name[grid.node_at_link_tail], out=out)

    return out


def map_mean_of_link_nodes_to_link(grid, var_name, out=None):
    """Map the mean of a link's nodes to the link.

    map_mean_of_link_nodes_to_link iterates across the grid and
    identifies the node values at both the "head" and "tail" of a given link.
    This function takes the sum of the two values of 'var_name' at both the
    "to" and "from" node. The average value of the two node values of
    'var_name' is then mapped to the link.

    Construction::

        map_mean_of_link_nodes_to_link(grid, var_name, out=None)

    Parameters
    ----------
    grid : ModelGrid
        A landlab ModelGrid.
    var_name : array or field name
        Values defined at nodes.
    out : ndarray, optional
        Buffer to place mapped values into or `None` to create a new array.

    Returns
    -------
    ndarray
        Mapped values at links.

    Examples
    --------
    >>> import numpy as np
    >>> from landlab.grid.mappers import map_mean_of_link_nodes_to_link
    >>> from landlab import RasterModelGrid

    >>> rmg = RasterModelGrid((3, 4))
    >>> rmg.at_node['z'] = np.array([ 0,  1,  2,  3,
    ...                               4,  5,  6,  7,
    ...                               8,  9, 10, 11])
    >>> map_mean_of_link_nodes_to_link(rmg, 'z')
    array([  0.5,   1.5,   2.5,   2. ,   3. ,   4. ,   5. ,   4.5,   5.5,
             6.5,   6. ,   7. ,   8. ,   9. ,   8.5,   9.5,  10.5])

    >>> values_at_links = rmg.empty(at='link')
    >>> rtn = map_mean_of_link_nodes_to_link(rmg, 'z', out=values_at_links)
    >>> values_at_links
    array([  0.5,   1.5,   2.5,   2. ,   3. ,   4. ,   5. ,   4.5,   5.5,
             6.5,   6. ,   7. ,   8. ,   9. ,   8.5,   9.5,  10.5])
    >>> rtn is values_at_links
    True
    """
    if out is None:
        out = grid.empty(at='link')

    if type(var_name) is str:
        var_name = grid.at_node[var_name]
    out[:] = 0.5 * (var_name[grid.node_at_link_head] +
                    var_name[grid.node_at_link_tail])

    return out


def map_value_at_min_node_to_link(grid, control_name, value_name, out=None):
    """
    Map the the value found in one node array to a link, based on the
    minimum value found in a second node field or array.

    map_value_at_min_node_to_link iterates across the grid and
    identifies the node values at both the "head" and "tail" of a given link.
    This function evaluates the value of 'control_name' at both the "to" and
    "from" node. The value of 'value_name' at the node with the minimum value
    of the two values of 'control_name' is then mapped to the link.

    Construction::

        map_value_at_min_node_to_link(grid, control_name, value_name,
                                      out=None)

    Parameters
    ----------
    grid : ModelGrid
        A landlab ModelGrid.
    control_name : array or field name
        Name of field defined at nodes or a node array that dictates which end
        of the link to draw values from.
    value_name : array or field name
        Name of field defined at nodes or  node array from which values are
        drawn, based on control_name.
    out : ndarray, optional
        Buffer to place mapped values into or `None` to create a new array.

    Returns
    -------
    ndarray
        Mapped values at links.

    Examples
    --------
    >>> import numpy as np
    >>> from landlab.grid.mappers import map_value_at_min_node_to_link
    >>> from landlab import RasterModelGrid

    >>> rmg = RasterModelGrid((3, 4))
    >>> _ = rmg.add_field('node', 'z',
    ...                   [[0, 1, 2, 3],
    ...                    [7, 6, 5, 4],
    ...                    [8, 9, 10, 11]])
    >>> _ = rmg.add_field('node', 'vals_to_map',
    ...                   [[0, 10, 20, 30],
    ...                    [70, 60, 50, 40],
    ...                    [80, 90, 100, 110]])
    >>> map_value_at_min_node_to_link(rmg, 'z', 'vals_to_map')
    array([   0.,   10.,   20.,    0.,   10.,   20.,   30.,   60.,   50.,
             40.,   70.,   60.,   50.,   40.,   80.,   90.,  100.])
    """
    if out is None:
        out = grid.empty(at='link')

    if type(control_name) is str:
        control_name = grid.at_node[control_name]
    if type(value_name) is str:
        value_name = grid.at_node[value_name]
    head_control = control_name[grid.node_at_link_head]
    tail_control = control_name[grid.node_at_link_tail]
    head_vals = value_name[grid.node_at_link_head]
    tail_vals = value_name[grid.node_at_link_tail]

    out[:] = np.where(tail_control < head_control, tail_vals, head_vals)
    return out


def map_value_at_max_node_to_link(grid, control_name, value_name, out=None):
    """
    Map the the value found in one node array to a link, based on the
    maximum value found in a second node field or array.

    map_value_at_max_node_to_link iterates across the grid and
    identifies the node values at both the "head" and "tail" of a given link.
    This function evaluates the value of 'control_name' at both the "to" and
    "from" node. The value of 'value_name' at the node with the maximum value
    of the two values of 'control_name' is then mapped to the link.

    Construction::

        map_value_at_max_node_to_link(grid, control_name, value_name,
                                      out=None)

    Parameters
    ----------
    grid : ModelGrid
        A landlab ModelGrid.
    control_name : array or field name
        Name of field defined at nodes or a node array that dictates which end
        of the link to draw values from.
    value_name : array or field name
        Name of field defined at nodes or  node array from which values are
        drawn, based on control_name.
    out : ndarray, optional
        Buffer to place mapped values into or `None` to create a new array.

    Returns
    -------
    ndarray
        Mapped values at links.

    Examples
    --------
    >>> import numpy as np
    >>> from landlab.grid.mappers import map_value_at_max_node_to_link
    >>> from landlab import RasterModelGrid

    >>> rmg = RasterModelGrid((3, 4))
    >>> _ = rmg.add_field('node', 'z',
    ...                   [[0, 1, 2, 3],
    ...                    [7, 6, 5, 4],
    ...                    [8, 9, 10, 11]])
    >>> _ = rmg.add_field('node', 'vals_to_map',
    ...                   [[0, 10, 20, 30],
    ...                    [70, 60, 50, 40],
    ...                    [80, 90, 100, 110]])
    >>> map_value_at_max_node_to_link(rmg, 'z', 'vals_to_map')
    array([  10.,   20.,   30.,   70.,   60.,   50.,   40.,   70.,   60.,
             50.,   80.,   90.,  100.,  110.,   90.,  100.,  110.])
    """
    if out is None:
        out = grid.empty(at='link')

    if type(control_name) is str:
        control_name = grid.at_node[control_name]
    if type(value_name) is str:
        value_name = grid.at_node[value_name]
    head_control = control_name[grid.node_at_link_head]
    tail_control = control_name[grid.node_at_link_tail]
    head_vals = value_name[grid.node_at_link_head]
    tail_vals = value_name[grid.node_at_link_tail]

    out[:] = np.where(tail_control > head_control, tail_vals, head_vals)
    return out


def map_node_to_cell(grid, var_name, out=None):
    """Map values for nodes to cells.

    map_node_to_cell iterates across the grid and
    identifies the all node values of 'var_name'.

    This function takes node values of 'var_name' and mapes that value to the
    corresponding cell area for each node.

    Construction::

        map_node_to_cell(grid, var_name, out=None)

    Parameters
    ----------
    grid : ModelGrid
        A landlab ModelGrid.
    var_name : array or field name
        Values defined at nodes.
    out : ndarray, optional
        Buffer to place mapped values into or `None` to create a new array.

    Returns
    -------
    ndarray
        Mapped values at cells.

    Examples
    --------
    >>> import numpy as np
    >>> from landlab.grid.mappers import map_node_to_cell
    >>> from landlab import RasterModelGrid

    >>> rmg = RasterModelGrid((3, 4))
    >>> _ = rmg.add_field('node', 'z', np.arange(12.))
    >>> map_node_to_cell(rmg, 'z')
    array([ 5.,  6.])

    >>> values_at_cells = rmg.empty(at='cell')
    >>> rtn = map_node_to_cell(rmg, 'z', out=values_at_cells)
    >>> values_at_cells
    array([ 5.,  6.])
    >>> rtn is values_at_cells
    True
    """
    if out is None:
        out = grid.empty(at='cell')

    if type(var_name) is str:
        var_name = grid.at_node[var_name]
    out[:] = var_name[grid.node_at_cell]

    return out


def map_min_of_node_links_to_node(grid, var_name, out=None):
    """Map the minimum value of a nodes' links to the node.

    map_min_of_node_links_to_node iterates across the grid and
    identifies the link values at each link connected to  a node.
    This function finds the minimum value of 'var_name' of each set
    of links, and then maps this value to the node. Note no attempt is made
    to honor the directionality of the links.

    Construction::

        map_min_of_node_links_to_node(grid, var_name, out=None)

    Parameters
    ----------
    grid : ModelGrid
        A landlab ModelGrid.
    var_name : array or field name
        Values defined at links.
    out : ndarray, optional
        Buffer to place mapped values into or `None` to create a new array.

    Returns
    -------
    ndarray
        Mapped values at nodes.

    Examples
    --------
    >>> import numpy as np
    >>> from landlab.grid.mappers import map_min_of_node_links_to_node
    >>> from landlab import RasterModelGrid

    >>> rmg = RasterModelGrid((3, 4))
    >>> rmg.at_link['grad'] = np.arange(rmg.number_of_links)
    >>> map_min_of_node_links_to_node(rmg, 'grad')
    array([  0.,   0.,   1.,   2.,
             3.,   4.,   5.,   6.,
            10.,  11.,  12.,  13.])

    >>> values_at_nodes = rmg.add_empty('node', 'z')
    >>> rtn = map_min_of_node_links_to_node(rmg, 'grad', out=values_at_nodes)
    >>> values_at_nodes
    array([  0.,   0.,   1.,   2.,
             3.,   4.,   5.,   6.,
            10.,  11.,  12.,  13.])
    >>> rtn is values_at_nodes
    True
    """
    if out is None:
        out = grid.empty(at='node')

    values_at_linksX = np.empty(grid.number_of_links+1, dtype=float)
    values_at_linksX[-1] = np.finfo(dtype=float).max
    if type(var_name) is str:
        values_at_linksX[:-1] = grid.at_link[var_name]
    else:
        values_at_linksX[:-1] = var_name
    np.amin(values_at_linksX[grid.links_at_node], axis=1, out=out)

    return out


def map_max_of_node_links_to_node(grid, var_name, out=None):
    """Map the maximum value of a nodes' links to the node.

    map_max_of_node_links_to_node iterates across the grid and
    identifies the link values at each link connected to  a node.
    This function finds the maximum value of 'var_name' of each set
    of links, and then maps this value to the node. Note no attempt is made
    to honor the directionality of the links.

    Construction::

        map_max_of_node_links_to_node(grid, var_name, out=None)

    Parameters
    ----------
    grid : ModelGrid
        A landlab ModelGrid.
    var_name : array or field name
        Values defined at links.
    out : ndarray, optional
        Buffer to place mapped values into or `None` to create a new array.

    Returns
    -------
    ndarray
        Mapped values at nodes.

    Examples
    --------
    >>> import numpy as np
    >>> from landlab.grid.mappers import map_max_of_node_links_to_node
    >>> from landlab import RasterModelGrid

    >>> rmg = RasterModelGrid((3, 4))
    >>> rmg.at_link['grad'] = np.arange(rmg.number_of_links)
    >>> map_max_of_node_links_to_node(rmg, 'grad')
    array([  3.,   4.,   5.,   6.,
            10.,  11.,  12.,  13.,
            14.,  15.,  16.,  16.])

    >>> values_at_nodes = rmg.add_empty('node', 'z')
    >>> rtn = map_max_of_node_links_to_node(rmg, 'grad', out=values_at_nodes)
    >>> values_at_nodes
    array([  3.,   4.,   5.,   6.,
            10.,  11.,  12.,  13.,
            14.,  15.,  16.,  16.])
    >>> rtn is values_at_nodes
    True
    """
    if out is None:
        out = grid.empty(at='node')

    values_at_linksX = np.empty(grid.number_of_links+1, dtype=float)
    values_at_linksX[-1] = np.finfo(dtype=float).min
    if type(var_name) is str:
        values_at_linksX[:-1] = grid.at_link[var_name]
    else:
        values_at_linksX[:-1] = var_name
    np.amax(values_at_linksX[grid.links_at_node], axis=1, out=out)

    return out


def map_upwind_node_link_max_to_node(grid, var_name, out=None):
    """
    Map the largest magnitude of the links bringing flux into the node to the
    node.

    map_upwind_node_link_max_to_node iterates across the grid and identifies
    the link values at each link connected to a node. It then uses the
    link_dirs_at_node data structure to identify links bringing flux into the
    node, then maps the maximum magnitude of 'var_name' found on these links
    onto the node. If no upwind link is found, the value will be recorded as
    zero.

    Construction::

        map_upwind_node_link_max_to_node(grid, var_name, out=None)

    Parameters
    ----------
    grid : ModelGrid
        A landlab ModelGrid.
    var_name : array or field name
        Values defined at links.
    out : ndarray, optional
        Buffer to place mapped values into or `None` to create a new array.

    Returns
    -------
    ndarray
        Mapped values at nodes.

    Examples
    --------
    >>> import numpy as np
    >>> from landlab.grid.mappers import map_upwind_node_link_max_to_node
    >>> from landlab import RasterModelGrid

    >>> rmg = RasterModelGrid((3, 4))
    >>> rmg.at_link['grad'] = np.array([-1., -2., -1.,
    ...                                  0.,  0.,  0.,  0.,
    ...                                 -1., -2., -1.,
    ...                                  0.,  0.,  0.,  0.,
    ...                                 -1., -2., -1.])
    >>> map_upwind_node_link_max_to_node(rmg, 'grad')
    array([ 0.,  1.,  2.,  1.,
            0.,  1.,  2.,  1.,
            0.,  1.,  2.,  1.])

    >>> values_at_nodes = rmg.add_empty('node', 'z')
    >>> rtn = map_upwind_node_link_max_to_node(rmg, 'grad',
    ...                                        out=values_at_nodes)
    >>> values_at_nodes
    array([ 0.,  1.,  2.,  1.,
            0.,  1.,  2.,  1.,
            0.,  1.,  2.,  1.])
    >>> rtn is values_at_nodes
    True
    """
    if out is None:
        out = grid.empty(at='node')

    if type(var_name) is str:
        var_name = grid.at_link[var_name]
    values_at_links = var_name[grid.links_at_node] * grid.link_dirs_at_node
    # this procedure makes incoming links NEGATIVE
    np.amax(-values_at_links, axis=1, out=out)

    return out


def map_downwind_node_link_max_to_node(grid, var_name, out=None):
    """
    Map the largest magnitude of the links carrying flux from the node to the
    node.

    map_downwind_node_link_max_to_node iterates across the grid and identifies
    the link values at each link connected to a node. It then uses the
    link_dirs_at_node data structure to identify links carrying flux out of the
    node, then maps the maximum magnitude of 'var_name' found on these links
    onto the node. If no downwind link is found, the value will be recorded as
    zero.

    Construction::

        map_downwind_node_link_max_to_node(grid, var_name, out=None)

    Parameters
    ----------
    grid : ModelGrid
        A landlab ModelGrid.
    var_name : array or field name
        Values defined at links.
    out : ndarray, optional
        Buffer to place mapped values into or `None` to create a new array.

    Returns
    -------
    ndarray
        Mapped values at nodes.

    Examples
    --------
    >>> import numpy as np
    >>> from landlab.grid.mappers import map_downwind_node_link_max_to_node
    >>> from landlab import RasterModelGrid

    >>> rmg = RasterModelGrid((3, 4))
    >>> rmg.at_link['grad'] = np.array([-1., -2., -1.,
    ...                                  0.,  0.,  0.,  0.,
    ...                                 -1., -2., -1.,
    ...                                  0.,  0.,  0.,  0.,
    ...                                 -1., -2., -1.])
    >>> map_downwind_node_link_max_to_node(rmg, 'grad')
    array([ 1.,  2.,  1.,  0.,
            1.,  2.,  1.,  0.,
            1.,  2.,  1.,  0.])

    >>> values_at_nodes = rmg.add_empty('node', 'z')
    >>> rtn = map_downwind_node_link_max_to_node(rmg, 'grad',
    ...                                        out=values_at_nodes)
    >>> values_at_nodes
    array([ 1.,  2.,  1.,  0.,
            1.,  2.,  1.,  0.,
            1.,  2.,  1.,  0.])
    >>> rtn is values_at_nodes
    True
    """
    if out is None:
        out = grid.empty(at='node')

    if type(var_name) is str:
        var_name = grid.at_link[var_name]
    values_at_links = var_name[grid.links_at_node] * grid.link_dirs_at_node
    # this procedure makes incoming links NEGATIVE
    steepest_links_at_node = np.amax(values_at_links, axis=1)
    np.fabs(steepest_links_at_node, out=out)

    return out


def map_upwind_node_link_mean_to_node(grid, var_name, out=None):
    """
    Map the mean magnitude of the links bringing flux into the node to the
    node.

    map_upwind_node_link_mean_to_node iterates across the grid and identifies
    the link values at each link connected to a node. It then uses the
    link_dirs_at_node data structure to identify links bringing flux into the
    node, then maps the mean magnitude of 'var_name' found on these links
    onto the node. Links with zero values are not included in the means,
    and zeros are returned if no upwind links are found.

    Construction::

        map_upwind_node_link_mean_to_node(grid, var_name, out=None)

    Parameters
    ----------
    grid : ModelGrid
        A landlab ModelGrid.
    var_name : array or field name
        Values defined at links.
    out : ndarray, optional
        Buffer to place mapped values into or `None` to create a new array.

    Returns
    -------
    ndarray
        Mapped values at nodes.

    Examples
    --------
    >>> import numpy as np
    >>> from landlab.grid.mappers import map_upwind_node_link_mean_to_node
    >>> from landlab import RasterModelGrid

    >>> rmg = RasterModelGrid((3, 4))
    >>> rmg.at_link['grad'] = np.array([-1., -2., -1.,
    ...                                 -2., -3., -4., -5.,
    ...                                 -1., -2., -1.,
    ...                                 -1., -2., -3., -4.,
    ...                                 -1., -2., -1.])
    >>> map_upwind_node_link_mean_to_node(rmg, 'grad')
    array([ 0. ,  1. ,  2. ,  1. ,
            2. ,  2. ,  3. ,  3. ,
            1. ,  1.5,  2.5,  2.5])

    >>> values_at_nodes = rmg.add_empty('node', 'z')
    >>> rtn = map_upwind_node_link_mean_to_node(rmg, 'grad',
    ...                                         out=values_at_nodes)
    >>> values_at_nodes
    array([ 0. ,  1. ,  2. ,  1. ,
            2. ,  2. ,  3. ,  3. ,
            1. ,  1.5,  2.5,  2.5])
    >>> rtn is values_at_nodes
    True
    """
    if out is None:
        out = grid.empty(at='node')

    if type(var_name) is str:
        var_name = grid.at_link[var_name]
    values_at_links = var_name[grid.links_at_node] * grid.link_dirs_at_node
    # this procedure makes incoming links NEGATIVE
    vals_in_positive = -values_at_links
    vals_above_zero = vals_in_positive > 0.
    total_vals = np.sum(vals_in_positive*vals_above_zero, axis=1)
    link_count = np.sum(vals_above_zero, axis=1)
    np.divide(total_vals, link_count, out=out)
    out[np.isnan(out)] = 0.

    return out


def map_downwind_node_link_mean_to_node(grid, var_name, out=None):
    """
    Map the mean magnitude of the links carrying flux out of the node to the
    node.

    map_downwind_node_link_mean_to_node iterates across the grid and identifies
    the link values at each link connected to a node. It then uses the
    link_dirs_at_node data structure to identify links carrying flux out of the
    node, then maps the mean magnitude of 'var_name' found on these links
    onto the node. Links with zero values are not included in the means,
    and zeros are returned if no upwind links are found.

    Construction::

        map_downwind_node_link_mean_to_node(grid, var_name, out=None)

    Parameters
    ----------
    grid : ModelGrid
        A landlab ModelGrid.
    var_name : array or field name
        Values defined at links.
    out : ndarray, optional
        Buffer to place mapped values into or `None` to create a new array.

    Returns
    -------
    ndarray
        Mapped values at nodes.

    Examples
    --------
    >>> import numpy as np
    >>> from landlab.grid.mappers import map_downwind_node_link_mean_to_node
    >>> from landlab import RasterModelGrid

    >>> rmg = RasterModelGrid((3, 4))
    >>> rmg.at_link['grad'] = np.array([-1., -2., -1.,
    ...                                 -2., -3., -4., -5.,
    ...                                 -1., -2., -1.,
    ...                                 -1., -2., -3., -4.,
    ...                                 -1., -2., -1.])
    >>> map_downwind_node_link_mean_to_node(rmg, 'grad')
    array([ 1.5,  2.5,  2.5,  5. ,
            1. ,  2. ,  2. ,  4. ,
            1. ,  2. ,  1. ,  0. ])

    >>> values_at_nodes = rmg.add_empty('node', 'z')
    >>> rtn = map_downwind_node_link_mean_to_node(rmg, 'grad',
    ...                                         out=values_at_nodes)
    >>> values_at_nodes
    array([ 1.5,  2.5,  2.5,  5. ,
            1. ,  2. ,  2. ,  4. ,
            1. ,  2. ,  1. ,  0. ])
    >>> rtn is values_at_nodes
    True
    """
    if out is None:
        out = grid.empty(at='node')

    if type(var_name) is str:
        var_name = grid.at_link[var_name]
    values_at_links = var_name[grid.links_at_node] * grid.link_dirs_at_node
    # this procedure makes incoming links NEGATIVE
    vals_in_positive = values_at_links
    vals_above_zero = vals_in_positive > 0.
    total_vals = np.sum(vals_in_positive*vals_above_zero, axis=1)
    link_count = np.sum(vals_above_zero, axis=1)
    np.divide(total_vals, link_count, out=out)
    out[np.isnan(out)] = 0.

    return out


def map_value_at_upwind_node_link_max_to_node(grid, control_name,
                                              value_name, out=None):
    """
    Map the the value found in one link array to a node, based on the
    largest magnitude value of links bringing fluxes into the node,
    found in a second node array or field.

    map_upwind_node_link_max_to_node iterates across the grid and identifies
    the link control_values at each link connected to a node. It then uses the
    link_dirs_at_node data structure to identify links bringing flux into the
    node, then identifies the link with the maximum magnitude. The value of the
    second field 'value_name' at these links is then mapped onto the node.
    If no upwind link is found, the value will be recorded as zero.

    Construction::

        map_value_at_upwind_node_link_max_to_node(grid, control_name,
                                                  value_name, out=None)

    Parameters
    ----------
    grid : ModelGrid
        A landlab ModelGrid.
    control_name : array or field name
        Values defined at nodes that dictate which end of the link to
        draw values from.
    value_name : array or field name
        Values defined at nodes from which values are drawn, based on
        control_name.
    out : ndarray, optional
        Buffer to place mapped values into or `None` to create a new array.

    Returns
    -------
    ndarray
        Mapped values at nodes.

    Examples
    --------
    >>> import numpy as np
    >>> from landlab.grid.mappers import map_value_at_upwind_node_link_max_to_node
    >>> from landlab import RasterModelGrid

    >>> rmg = RasterModelGrid((3, 4))
    >>> rmg.at_link['grad'] = np.array([-1., -2., -1.,
    ...                                  0.,  0.,  0.,  0.,
    ...                                 -1., -2., -1.,
    ...                                  0.,  0.,  0.,  0.,
    ...                                 -1., -2., -1.])
    >>> rmg.at_link['vals'] = np.arange(rmg.number_of_links, dtype=float)
    >>> map_value_at_upwind_node_link_max_to_node(rmg, 'grad', 'vals')
    array([  0.,   0.,   1.,   2.,
             0.,   7.,   8.,   9.,
             0.,  14.,  15.,  16.])

    >>> values_at_nodes = rmg.add_empty('node', 'z')
    >>> rtn = map_value_at_upwind_node_link_max_to_node(rmg, 'grad', 'vals',
    ...                                                 out=values_at_nodes)
    >>> values_at_nodes
    array([  0.,   0.,   1.,   2.,
             0.,   7.,   8.,   9.,
             0.,  14.,  15.,  16.])
    >>> rtn is values_at_nodes
    True
    """
    if out is None:
        out = grid.empty(at='node')

    if type(control_name) is str:
        control_name = grid.at_link[control_name]
    if type(value_name) is str:
        value_name = grid.at_link[value_name]
    values_at_nodes = control_name[grid.links_at_node] * grid.link_dirs_at_node
    # this procedure makes incoming links NEGATIVE
    which_link = np.argmax(-values_at_nodes, axis=1)
    invalid_links = values_at_nodes >= 0.
    link_vals_without_invalids = value_name[grid.links_at_node]
    link_vals_without_invalids[invalid_links] = 0.
    out[:] = link_vals_without_invalids[np.arange(grid.number_of_nodes),
                                        which_link]

    return out


def map_value_at_downwind_node_link_max_to_node(grid, control_name,
                                                value_name, out=None):
    """
    Map the the value found in one link array to a node, based on the
    largest magnitude value of links carrying fluxes out of the node,
    found in a second node array or field.

    map_downwind_node_link_max_to_node iterates across the grid and identifies
    the link control_values at each link connected to a node. It then uses the
    link_dirs_at_node data structure to identify links carrying flux out of the
    node, then identifies the link with the maximum magnitude. The value of the
    second field 'value_name' at these links is then mapped onto the node.
    If no downwind link is found, the value will be recorded as zero.

    Construction::

        map_value_at_downwind_node_link_max_to_node(grid, control_name,
                                                    value_name, out=None)

    Parameters
    ----------
    grid : ModelGrid
        A landlab ModelGrid.
    control_name : array or field name
        Values defined at nodes that dictate which end of the link to
        draw values from.
    value_name : array or field name
        Values defined at nodes from which values are drawn, based on
        control_name.
    out : ndarray, optional
        Buffer to place mapped values into or `None` to create a new array.

    Returns
    -------
    ndarray
        Mapped values at nodes.

    Examples
    --------
    >>> import numpy as np
    >>> from landlab.grid.mappers import map_value_at_downwind_node_link_max_to_node
    >>> from landlab import RasterModelGrid

    >>> rmg = RasterModelGrid((3, 4))
    >>> rmg.at_link['grad'] = np.array([-1., -2., -1.,
    ...                                  0.,  0.,  0.,  0.,
    ...                                 -1., -2., -1.,
    ...                                  0.,  0.,  0.,  0.,
    ...                                 -1., -2., -1.])
    >>> rmg.at_link['vals'] = np.arange(rmg.number_of_links, dtype=float)
    >>> map_value_at_downwind_node_link_max_to_node(rmg, 'grad', 'vals')
    array([  0.,   1.,   2.,   0.,
             7.,   8.,   9.,   0.,
            14.,  15.,  16.,   0.])

    >>> values_at_nodes = rmg.add_empty('node', 'z')
    >>> rtn = map_value_at_downwind_node_link_max_to_node(rmg, 'grad', 'vals',
    ...                                                   out=values_at_nodes)
    >>> values_at_nodes
    array([  0.,   1.,   2.,   0.,
             7.,   8.,   9.,   0.,
            14.,  15.,  16.,   0.])
    >>> rtn is values_at_nodes
    True
    """
    if out is None:
        out = grid.empty(at='node')

    if type(control_name) is str:
        control_name = grid.at_link[control_name]
    if type(value_name) is str:
        value_name = grid.at_link[value_name]
    values_at_nodes = control_name[grid.links_at_node] * grid.link_dirs_at_node
    # this procedure makes incoming links NEGATIVE
    which_link = np.argmax(values_at_nodes, axis=1)
    invalid_links = values_at_nodes <= 0.
    link_vals_without_invalids = value_name[grid.links_at_node]
    link_vals_without_invalids[invalid_links] = 0.
    out[:] = link_vals_without_invalids[np.arange(grid.number_of_nodes),
                                        which_link]

    return out


def dummy_func_to_demonstrate_docstring_modification(grid, some_arg):
    """A dummy function to demonstrate automated docstring changes.

    Construction::

        dummy_func_to_demonstrate_docstring_modification(grid, some_arg)

    Parameters
    ----------
    grid : ModelGrid
        A Landlab modelgrid.
    some_arg : whatever
        A dummy argument.

    Examples
    --------
    ...
    """
    pass



#!/bin/env python

import base64
import numpy as np


class EncoderError(Exception):
    pass


class UnknownEncoderError(EncoderError):

    def __init__(self, name):
        self._name

    def __str__(self):
        return '%s: Unknown encoder' % self._name


class EncoderInterface(object):

    def encode(self, array):
        pass


class AsciiEncoder(object):

    def encode(self, array):
        try:
            return ' '.join([str(val) for val in array.flatten()])
        except AttributeError:
            return ' '.join([str(val) for val in array])


class RawEncoder(object):

    def encode(self, array):
        try:
            as_str = array.tostring()
        except AttributeError:
            as_str = np.array(array).tostring()
        block_size = np.array(len(as_str), dtype=np.int32).tostring()
        return block_size + as_str


class Base64Encoder(object):

    def encode(self, array):
        try:
            as_str = array.tostring()
        except AttributeError:
            as_str = np.array(array).tostring()
        as_str = base64.b64encode(as_str)
        block_size = base64.b64encode(np.array(len(as_str),
                                               dtype=np.int32).tostring())
        return block_size + as_str

    def decode(self, array):
        pass


_ENCODERS = {
    'ascii': AsciiEncoder(),
    'raw': RawEncoder(), 'base64': Base64Encoder()
}


def encode(array, encoding='ascii'):
    return _ENCODERS[encoding].encode(array)


def decode(array, encoding='raw'):
    return _ENCODERS[encoding].decode(array)



#! /usr/env/python
"""The landlab flow accumulation module.

A python flow accumulation module. It is designed to be general, and to
operate across multiple grids and multiple flow direction patterns. However,
at the moment, only a steepest descent (single path) routing scheme is
implemented.

Notes
-----
There remain some outstanding issues with the handling of boundary cells,
which this component has inherited from `flow_routing_D81.
"""
# Created DEJH, 8/2013
from __future__ import print_function

from six.moves import range

import warnings

import numpy as np

from landlab import CLOSED_BOUNDARY


class AccumFlow(object):

    """Flow accumulation module.

    This class allows the routing of flow around a landscape according to a
    previously calculated flow direction vector. It is not sensitive to grid
    type. It will eventually be able to work with discharges which are split
    across more than one node, but at the moment, assumes a single line of
    descent for a given node.
    """

    def __init__(self, grid):
        self._grid = grid

        # Prefilled with zeros, size of WHOLE grid + 1, to allow -1 ids
        self._flow_accum_by_area = np.empty(grid.number_of_nodes + 1,
                                            dtype=float)
        self._flow_accum_by_area[-1] = 0.

    @property
    def grid(self):
        return self._grid

    def calc_flowacc(self, z, flowdirs):
        """Calculate flow accumulation.

        Parameters
        ----------
        z : ndarray
            Elevations at nodes.
        flowdirs : ndarray
            Downstream neighbor node.

        Returns
        -------
        ndarray
            Total upstream area at each node.

        Examples
        --------
        >>> import numpy as np
        >>> from landlab import RasterModelGrid

        >>> grid = RasterModelGrid((4, 5), spacing=(3, 4))
        >>> z = np.array([3., 3., 0., 3., 3.,
        ...               3., 2., 1., 2., 3.,
        ...               3., 2., 2., 2., 3.,
        ...               3., 3., 3., 3., 3.])

        First calculate the flow directions.

        >>> from landlab.components.flow_routing import RouteFlowD8
        >>> flow_router = RouteFlowD8(len(z))
        >>> flowdirs, _ = flow_router.calc_flowdirs(grid ,z)
        >>> flowdirs
        array([ 6,  7, -1,  7,  8,
                6,  2,  2,  2,  8,
               11,  7,  7,  7, 13,
               11, 11, 12, 13, 13])

        Then calculate the drainage areas.

        >>> from landlab.components.flow_accum import AccumFlow
        >>> accumulator = AccumFlow(grid)
        >>> accumulator.calc_flowacc(z, flowdirs)
        array([  0.,   0.,  72.,   0.,   0.,
                 0.,  12.,  48.,  12.,   0.,
                 0.,  12.,  12.,  12.,   0.,
                 0.,   0.,   0.,   0.,   0.])
        """
        nodes = np.where(self.grid.status_at_node != CLOSED_BOUNDARY)[0]

        # Cell areas for nodes without cells is 0.
        self._flow_accum_by_area[nodes] = self.grid.cell_area_at_node[nodes]

        # Perform test to see if the flowdir data is a single vector, or
        # multidimensional, here. Several ways possible: 1. Is the vector
        # multidimensional?, e.g., try: data.flowdirs.shape[1] 2. set a flag
        # in flowdir.

        try:
            # Descending order.
            height_order_nodes = np.argsort(z[nodes])[::-1]
        except:
            warnings.warn("Cells could not be sorted by elevation. Does the "
                          "data object contain the elevation vector?")

        try:
            sorted_flowdirs = flowdirs[nodes][height_order_nodes]
        except:
            warnings.warn("Flow directions could not be sorted by elevation. "
                          "Does the data object contain the flow direction "
                          "vector?")

        ## Inefficient Python code.
        for i in range(len(sorted_flowdirs)):
            iter_height_order = height_order_nodes[i]
            iter_sorted_fldirs = sorted_flowdirs[i]
            self._flow_accum_by_area[iter_sorted_fldirs] += (
                self._flow_accum_by_area[nodes][iter_height_order])

        return self._flow_accum_by_area[:-1]



"""
Unit tests for landlab.components.radiation.radiation
"""
from nose.tools import assert_equal, assert_true, assert_raises, with_setup
from numpy.testing import assert_array_almost_equal
try:
    from nose.tools import assert_is_instance
except ImportError:
    from landlab.testing.tools import assert_is_instance
import numpy as np

from landlab import RasterModelGrid
from landlab.components.radiation.radiation import Radiation


(_SHAPE, _SPACING, _ORIGIN) = ((20, 20), (10e0, 10e0), (0., 0.))
_ARGS = (_SHAPE, _SPACING, _ORIGIN)


def setup_grid():
    from landlab import RasterModelGrid
    grid = RasterModelGrid((20, 20), spacing=10e0)
    rad = Radiation(grid)
    globals().update({
        'rad': Radiation(grid)
    })


@with_setup(setup_grid)
def test_name():
    assert_equal(rad.name, 'Radiation')


@with_setup(setup_grid)
def test_input_var_names():
    assert_equal(rad.input_var_names,
                 ('topographic__elevation',))


@with_setup(setup_grid)
def test_output_var_names():
    assert_equal(sorted(rad.output_var_names),
                 ['radiation__incoming_shortwave',
                  'radiation__net_shortwave',
                  'radiation__ratio_to_flat_surface'])


@with_setup(setup_grid)
def test_var_units():
    assert_equal(set(rad.input_var_names) |
                 set(rad.output_var_names),
                 set(dict(rad.units).keys()))

    assert_equal(rad.var_units('topographic__elevation'), 'm')
    assert_equal(rad.var_units('radiation__incoming_shortwave'), 'W/m^2')
    assert_equal(rad.var_units('radiation__net_shortwave'), 'W/m^2')
    assert_equal(rad.var_units('radiation__ratio_to_flat_surface'), 'None')


@with_setup(setup_grid)
def test_grid_shape():
    assert_equal(rad.grid.number_of_node_rows, _SHAPE[0])
    assert_equal(rad.grid.number_of_node_columns, _SHAPE[1])


@with_setup(setup_grid)
def test_grid_x_extent():
    assert_equal(rad.grid.extent[1], (_SHAPE[1] - 1) * _SPACING[1])


@with_setup(setup_grid)
def test_grid_y_extent():
    assert_equal(rad.grid.extent[0], (_SHAPE[0] - 1) * _SPACING[0])


@with_setup(setup_grid)
def test_field_getters():
    for name in rad.grid['node']:
        field = rad.grid['node'][name]
        assert_is_instance(field, np.ndarray)
        assert_equal(field.shape,
                     (rad.grid.number_of_node_rows *
                      rad.grid.number_of_node_columns, ))
                      
    for name in rad.grid['cell']:
        field = rad.grid['cell'][name]
        assert_is_instance(field, np.ndarray)
        assert_equal(field.shape,
                     (rad.grid.number_of_cell_rows *
                      rad.grid.number_of_cell_columns, ))

    assert_raises(KeyError, lambda: rad.grid['not_a_var_name'])


@with_setup(setup_grid)
def test_field_initialized_to_zero():
    for name in rad.grid['node']:
        field = rad.grid['node'][name]
        assert_array_almost_equal(field, np.zeros(rad.grid.number_of_nodes))
    for name in rad.grid['cell']:
        if name == 'Slope' or name == 'Aspect':
            continue
        field = rad.grid['cell'][name]
        assert_array_almost_equal(field, np.zeros(rad.grid.number_of_cells))


# -*- coding: utf-8 -*-
"""
This is a Landlab wrapper for A Wickert's gFlex flexure model (Wickert et al.,
submitted to Geoscientific Model Development). The most up-to-date version of
his code can be found at github.com/awickert/gFlex.

This Landlab wrapper will use a snapshot of that code, which YOU need to
install on your own machine.
A stable snapshot of gFlex is hosted on PyPI, which is the recommended version
to install.
If you have pip (the Python package install tool), simply run
'pip install gFlex' from a command prompt.
Alternatively, you can download and unpack the code (from github, or with PyPI,
pypi.python.org/pypi/gFlex/), then run 'python setup.py install'.

Created on Thu Feb 19 18:47:11 2015

@author: daniel.hobley (SiccarPoint @Github)

...following AW's run_in_script_2D.py.
"""
from __future__ import print_function

import numpy as np
import inspect
from landlab import RasterModelGrid, Component
from landlab import ModelParameterDictionary
from landlab import FieldError

try:
    import gflex
except ImportError:
    import warnings, sys

    warnings.warn("gFlex not installed.")
    print("""
To use the gFlex component you must have gFlex installed on your machine.
For installation instructions see gFlex on GitHub:

  https://github.com/awickert/gFlex
          """.strip(), file=sys.stderr)


class gFlex(Component):
    """
    This is a Landlab wrapper for A Wickert's gFlex flexure model (Wickert et
    al., 2016, Geoscientific Model Development). The most up-to-date version of
    his code can be found at github.com/awickert/gFlex.

    This Landlab wrapper will use a snapshot of that code, which YOU need to
    install on your own machine.
    A stable snapshot of gFlex is hosted on PyPI, which is the recommended
    version to install.
    If you have pip (the Python package install tool), simply run
    'pip install gFlex' from a command prompt.
    Alternatively, you can download and unpack the code (from github, or with
    PyPI, pypi.python.org/pypi/gFlex/), then run 'python setup.py install'.

    Note that gFlex maintains its own internal version if the grid, but this
    should not affect performance.

    This component will modify the topographic__elevation field only if one
    already exists. Note that the gFlex component **demands lengths in
    meters**, including the grid dimensions.

    Parameters
    ----------
    Youngs_modulus : float
        Young's modulus for the lithosphere.
    Poissons_ratio : float
        Poisson's ratio for the lithosphere.
    rho_mantle : float (kg*m**-3)
        The density of the mantle.
    rho_fill : float (kg*m**-3)
        The density of the infilling material (air, water...)
    elastic_thickness : float (m)
        The elastic thickness of the lithosphere.
    BC_W, BC_E, BC_N, BC_S : {'0Displacement0Slope', '0Moment0Shear',
                              'Periodic'}
        The boundary condition status of each grid edge, following gFlex's
        definitions. Periodic boundaries must be paired (obviously).
    g : float (m*s**-2)
        The acceleration due to gravity.

    The component also recognises the gFlex specific parameters 'Method',
    'PlateSolutionType', 'Solver', and 'Quiet'. See the gFlex software
    documentation for more details.

    Examples
    --------

    NB: these tests are not actually run as our automated testing becomes
    confused if gFlex is not installed on the testing machine!

    >>> from landlab import RasterModelGrid
    >>> from landlab.components import gFlex
    >>> mg = RasterModelGrid((10, 10), 25000.)
    >>> z = mg.add_zeros('topographic__elevation', at='node', dtype=float)
    >>> stress = mg.add_zeros('surface_load__stress', at='node', dtype=float)
    >>> stress.view().reshape(mg.shape)[3:7, 3:7] += 1.e6
    >>> gf = gFlex(mg, BC_E='0Moment0Shear', BC_N='Periodic',
    ...            BC_S='Periodic') # doctest: +SKIP
    >>> gf.run_one_step() # doctest: +SKIP

    N-S profile across flexed plate:

    >>> z.reshape(mg.shape)[:, 5] # doctest: +SKIP
    array([-4.54872677, -4.6484927 , -4.82638669, -5.03001546, -5.15351385,
           -5.15351385, -5.03001546, -4.82638669, -4.6484927 , -4.54872677])

    W-E profile, noting the free BC to the east side:

    >>> z.reshape(mg.shape)[5, :] # doctest: +SKIP
    array([-0.43536739, -1.19197738, -2.164915  , -3.2388464 , -4.2607558 ,
           -5.15351385, -5.89373366, -6.50676947, -7.07880156, -7.63302576])
    """

    _name = 'gFlex'

    _input_var_names = ('surface_load__stress',)

    _output_var_names = ('lithosphere__vertical_displacement',
                         'topographic__elevation')

    _var_units = {'surface_load__stress': 'Pa',
                  'lithosphere__vertical_displacement': 'm',
                  'topographic__elevation': 'm'
                  }

    _var_mapping = {'surface_load__stress': 'node',
                    'lithosphere__vertical_displacement': 'node',
                    'topographic__elevation': 'node'
                    }

    _var_doc = {'surface_load__stress': (
                    'Magnitude of stress exerted by surface load'),
                'lithosphere__vertical_displacement':
                    ('Vertical deflection of the surface and of the ' +
                     'lithospheric plate'),
                'topographic__elevation': (
                    'Land surface topographic elevation; can ' +
                    'be overwritten in initialization'),
                }

    def __init__(self, grid, Youngs_modulus=6.5e11, Poissons_ratio=0.25,
                 rho_mantle=3300., rho_fill=0., elastic_thickness=35000.,
                 BC_W='0Displacement0Slope', BC_E='0Displacement0Slope',
                 BC_N='0Displacement0Slope', BC_S='0Displacement0Slope',
                 g=9.81, **kwds):
        """Constructor for Wickert's gFlex in Landlab."""
        assert RasterModelGrid in inspect.getmro(grid.__class__)
        BC_options = ('0Displacement0Slope', '0Moment0Shear', '0Slope0Shear',
                      'Periodic')

        # instantiate the module:
        self.flex = gflex.F2D()
        flex = self.flex

        # set up the grid variables:
        self._grid = grid
        flex.dx = grid.dx
        flex.dy = grid.dy

        # we assume these properties are fixed in this relatively
        # straightforward implementation, but they can still be set if you
        # want:
        try:
            flex.Method = kwds['Method']
        except KeyError:
            flex.Method = 'FD'
        try:
            flex.PlateSolutionType = kwds['PlateSolutionType']
        except KeyError:
            flex.PlateSolutionType = 'vWC1994'
        try:
            flex.Solver = kwds['Solver']
        except KeyError:
            flex.Solver = 'direct'
        try:
            quiet = kwds['Quiet']
        except KeyError:
            flex.Quiet = True
        else:
            flex.Quiet = bool(quiet)

        flex.E = float(Youngs_modulus)
        flex.nu = float(Poissons_ratio)
        flex.rho_m = float(rho_mantle)
        flex.rho_fill = float(rho_fill)
        flex.g = float(g)
        flex.BC_W = BC_W
        flex.BC_E = BC_E
        flex.BC_S = BC_S
        flex.BC_N = BC_N
        for i in (flex.BC_E, flex.BC_W, flex.BC_N, flex.BC_S):
            assert i in BC_options

        if BC_W == 'Periodic':
            assert BC_E == 'Periodic'
        if BC_E == 'Periodic':
            assert BC_W == 'Periodic'
        if BC_N == 'Periodic':
            assert BC_S == 'Periodic'
        if BC_S == 'Periodic':
            assert BC_N == 'Periodic'

        Te_in = elastic_thickness
        try:
            flex.Te = float(Te_in)
        except ValueError:
            try:
                flex.Te = grid.at_node[Te_in].view().reshape(grid.shape)
            except TypeError:
                flex.Te = Te_in.view().reshape(grid.shape)
            self._input_var_names.add(Te_in)
            self._output_var_names.add(Te_in)

        # set up the link between surface load stresses in the gFlex component
        # and the LL grid field:
        flex.qs = grid.at_node['surface_load__stress'].view().reshape(
            grid.shape)

        # create a holder for the "pre-flexure" state of the grid, to allow
        # updating of elevs:
        self.pre_flex = np.zeros(grid.number_of_nodes, dtype=float)

        # create the primary output field:
        self.grid.add_zeros('lithosphere__vertical_displacement', at='node',
                            dtype=float, noclobber=False)

    def flex_lithosphere(self, **kwds):
        """
        Executes (& finalizes, from the perspective of gFlex) the core method
        of gFlex. Note that flexure of the lithosphere proceeds to steady state
        in a single timestep.
        """
        # note kwds is redundant at the moment, but could be used subsequently
        # for dynamic control over params
        self.flex.qs = self.grid.at_node[
            'surface_load__stress'].view().reshape(self.grid.shape)
        self.flex.initialize()
        self.flex.run()
        self.flex.finalize()

        self.grid.at_node['lithosphere__vertical_displacement'][:] = \
            self.flex.w.view().ravel()

        try:
            self._grid.at_node['topographic__elevation']
            # a topo exists...
        except FieldError:
            pass
        else:
            topo_diff = self.grid.at_node[
                'lithosphere__vertical_displacement'] - self.pre_flex
            self.grid.at_node['topographic__elevation'] += topo_diff
            self.pre_flex += topo_diff

    def run_one_step(self, **kwds):
        """
        Flex the lithosphere to find its steady state form.

        The standardized run method for this component.

        Parameters
        ----------
        None
        """
        self.flex_lithosphere(**kwds)



from .flow_accumulation2 import AccumFlow
from .flow_accum_bw import (make_ordered_node_array,
                            find_drainage_area_and_discharge,
                            flow_accumulation)


__all__ = ['AccumFlow', 'make_ordered_node_array',
           'find_drainage_area_and_discharge', 'flow_accumulation', ]



# -*- coding: utf-8 -*-
from __future__ import print_function

import warnings

import numpy as np
from landlab import ModelParameterDictionary, CLOSED_BOUNDARY, Component

from landlab.core.model_parameter_dictionary import MissingKeyError, \
    ParameterValueError
from landlab.field.scalar_data_fields import FieldError
from landlab.grid.base import BAD_INDEX_VALUE
from landlab.utils.decorators import use_file_name_or_kwds
try:
    from .cfuncs import (erode_with_link_alpha_varthresh,
                         erode_with_link_alpha_fixthresh)
except ImportError:
    warnings.warn('Unable to import stream_power extension module.')
from copy import deepcopy as copy
UNDEFINED_INDEX = np.iinfo(np.int32).max


class StreamPowerEroder(Component):
    """Erode where channels are.

    Implemented as:

    .. math::
        E = K A^m S^n - sp_{crit},

    and if :math:`E < 0`, :math:`E = 0`.

    If ``use_W`` is declared and ``True``, the module instead implements:

    .. math::
        E = K A^m S^n / W - sp_{crit}

    DEJH Sept 2013, major modifications Sept 14 and May 16. This component
    now wraps Fastscape-style functionality under the hood.

    NB: If you want spatially or temporally variable runoff, pass the
    runoff values at each pixel to the flow router using the input argument
    *use_Q*.

    Construction::

        StreamPowerEroder(grid, K_sp=None, threshold_sp=0., sp_type='set_mn',
                          m_sp=0.5, n_sp=1., a_sp=None, b_sp=None, c_sp=None,
                          use_W=None, use_Q=None)

    Parameters
    ----------
    grid : ModelGrid
        A grid.
    K_sp : float, array, or field name
        K in the stream power equation (units vary with other parameters).
    threshold_sp : positive float, optional
        The threshold stream power, below which no erosion occurs. This
        threshold is assumed to be in "stream power" units, i.e., if
        sp_type is 'Shear_stress', the value should be tau**a.
    sp_type : {'set_mn', 'Total', 'Unit', 'Shear_stress'}
        Controls how the law is implemented. If 'set_mn', use the supplied
        values of m_sp and n_sp. Else, component will derive values of m and n
        from supplied values of a_sp, b_sp, and c_sp, following Whipple and
        Tucker:

        *  If ``'Total'``, ``m = a * c``, ``n = a``.
        *  If ``'Unit'``, ``m = a * c *(1 - b)``, ``n = a``.
        *  If ``'Shear_stress'``, ``m = 2 * a * c * (1 - b) / 3``,
           ``n = 2 * a / 3``.

    m_sp : float, optional
        m in the stream power equation (power on drainage area). Overridden if
        a_sp, b_sp, and c_sp are supplied.
    n_sp : float, optional, ~ 0.5<n_sp<4.
        n in the stream power equation (power on slope). Overridden if
        a_sp, b_sp, and c_sp are supplied.
    a_sp : float, optional
        The power on the SP/shear term to get the erosion rate; the "erosional
        process" term. Only used if sp_type is not 'set_mn'.
    b_sp : float, optional
        The power on discharge to get width; the "hydraulic geometry" term.
        Only used if sp_type in ('Unit', 'Shear_stress').
    c_sp : float, optional
        The power on area to get discharge; the "basin hydology" term. Only
        used if sp_type is not 'set_mn'.
    use_W : None, array, or field name, optional
        If not None, component will look for node-centered data describing
        channel width in grid.at_node[use_W] or if an array, will take the
        array as the channel widths. It will use the widths to implement
        incision ~ stream power per unit width. If sp_type is 'set_mn',
        follows the equation given above. If sp_type in ('Unit',
        'Shear_stress'), the width value will be implemented directly. W has no
        effect if sp_type is 'Total'.
    use_Q : None, array, or field name, optional
        If not None, the equation becomes E=K*Q**m*S**n. Effectively sets c=1
        in Wh&T's 1999 derivation, if you are setting m and n through a, b,
        and c.

    Examples
    --------
    >>> import numpy as np
    >>> from landlab import RasterModelGrid
    >>> from landlab import CLOSED_BOUNDARY, FIXED_VALUE_BOUNDARY
    >>> from landlab.components import FlowRouter
    >>> from landlab.components import StreamPowerEroder
    >>> mg = RasterModelGrid((5, 5), 10.)
    >>> z = np.array([7.,  7.,  7.,  7.,  7.,
    ...               7.,  5., 3.2,  6.,  7.,
    ...               7.,  2.,  3.,  5.,  7.,
    ...               7.,  1., 1.9,  4.,  7.,
    ...               7.,  0.,  7.,  7.,  7.])
    >>> z = mg.add_field('node', 'topographic__elevation', z)
    >>> fr = FlowRouter(mg)
    >>> sp = StreamPowerEroder(mg, K_sp=1.)
    >>> fr.run_one_step()
    >>> sp.run_one_step(dt=1.)
    >>> z  # doctest: +NORMALIZE_WHITESPACE
    array([ 7.        ,  7.        ,  7.        ,  7.        ,  7.        ,
            7.        ,  2.92996598,  2.02996598,  4.01498299,  7.        ,
            7.        ,  0.85993197,  1.87743897,  3.28268321,  7.        ,
            7.        ,  0.28989795,  0.85403051,  2.42701526,  7.        ,
            7.        ,  0.        ,  7.        ,  7.        ,  7.        ])

    >>> mg2 = RasterModelGrid((3, 7), 1.)
    >>> z = np.array(mg2.node_x**2.)
    >>> z = mg2.add_field('node', 'topographic__elevation', z)
    >>> mg2.status_at_node[mg2.nodes_at_left_edge] = FIXED_VALUE_BOUNDARY
    >>> mg2.status_at_node[mg2.nodes_at_top_edge] = CLOSED_BOUNDARY
    >>> mg2.status_at_node[mg2.nodes_at_bottom_edge] = CLOSED_BOUNDARY
    >>> mg2.status_at_node[mg2.nodes_at_right_edge] = CLOSED_BOUNDARY
    >>> fr2 = FlowRouter(mg2)
    >>> sp2 = StreamPowerEroder(mg2, K_sp=0.1, m_sp=0., n_sp=2.,
    ...                         threshold_sp=2.)
    >>> fr2.run_one_step()
    >>> sp2.run_one_step(dt=10.)
    >>> z.reshape((3, 7))[1, :]  # doctest: +NORMALIZE_WHITESPACE
    array([  0.        ,   1.        ,   4.        ,   8.52493781,
            13.29039716,  18.44367965,  36.        ])

    >>> mg3 = RasterModelGrid((5, 5), 2.)
    >>> z = mg.node_x/100.
    >>> z = mg3.add_field('node', 'topographic__elevation', z)
    >>> mg3.status_at_node[mg3.nodes_at_left_edge] = FIXED_VALUE_BOUNDARY
    >>> mg3.status_at_node[mg3.nodes_at_top_edge] = CLOSED_BOUNDARY
    >>> mg3.status_at_node[mg3.nodes_at_bottom_edge] = CLOSED_BOUNDARY
    >>> mg3.status_at_node[mg3.nodes_at_right_edge] = CLOSED_BOUNDARY
    >>> mg3.at_node['water__unit_flux_in'] = mg3.node_y
    >>> fr3 = FlowRouter(mg3)
    >>> Q = mg3.at_node['water__discharge']
    >>> sp3 = StreamPowerEroder(mg3, K_sp=1., sp_type='Unit', a_sp=1.,
    ...                         b_sp=0.5, c_sp=1., use_Q=Q)
    >>> fr3.run_one_step()
    >>> sp3.run_one_step(1.)
    >>> z
    array([ 0.        ,  0.1       ,  0.2       ,  0.3       ,  0.4       ,
            0.        ,  0.02898979,  0.0859932 ,  0.17463772,  0.4       ,
            0.        ,  0.02240092,  0.06879049,  0.14586033,  0.4       ,
            0.        ,  0.01907436,  0.05960337,  0.12929386,  0.4       ,
            0.        ,  0.1       ,  0.2       ,  0.3       ,  0.4       ])
    """

    _name = 'StreamPowerEroder'

    _input_var_names = (
        'topographic__elevation',
        'flow__link_to_receiver_node',
        'drainage_area',
        'flow__receiver_node',
        'flow__upstream_node_order',
        'topographic__steepest_slope'
    )

    _output_var_names = (
        'topographic__elevation',
    )

    _var_units = {
        'topographic__elevation': 'm',
        'drainage_area': 'm**2',
        'flow__link_to_receiver_node': '-',
        'flow__receiver_node': '-',
        'flow__upstream_node_order': '-',
        'topographic__steepest_slope': '-'
    }

    _var_mapping = {
        'topographic__elevation': 'node',
        'drainage_area': 'node',
        'flow__link_to_receiver_node': 'node',
        'flow__receiver_node': 'node',
        'flow__upstream_node_order': 'node',
        'topographic__steepest_slope': 'node'
    }

    _var_doc = {
        'topographic__elevation': 'Land surface topographic elevation',
        'drainage_area':
            "Upstream accumulated surface area contributing to the node's "
            "discharge",
        'flow__link_to_receiver_node':
            'ID of link downstream of each node, which carries the discharge',
        'flow__receiver_node':
            'Node array of receivers (node that receives flow from current '
            'node)',
        'flow__upstream_node_order':
            'Node array containing downstream-to-upstream ordered list of '
            'node IDs',
        'topographic__steepest_slope':
            'Node array of steepest *downhill* slopes'
    }

    @use_file_name_or_kwds
    def __init__(self, grid, K_sp=None, threshold_sp=0., sp_type='set_mn',
                 m_sp=0.5, n_sp=1., a_sp=None, b_sp=None, c_sp=None,
                 use_W=None, use_Q=None, **kwds):
        self._grid = grid
        self.fraction_gradient_change = 1.
        self.link_S_with_trailing_blank = np.zeros(grid.number_of_links+1)
        # ^needs to be filled with values in execution
        self.count_active_links = np.zeros_like(
            self.link_S_with_trailing_blank, dtype=int)
        self.count_active_links[:-1] = 1

        active_nodes = grid.status_at_node != 4
        # self._K_unit_time = np.empty(active_nodes.sum(), dtype=float)
        self._K_unit_time = self.grid.zeros('node', dtype=float)
        self.use_K = False  # grandfathered in; only if K_sp == 'array'
        if type(K_sp) is np.ndarray:
            self._K_unit_time[:] = K_sp
        else:
            try:
                self._K_unit_time.fill(K_sp)
            except ValueError:  # could not cast => was a str
                if K_sp == 'array':
                    self.use_K = True
                else:
                    self._K_unit_time = grid.at_node[K_sp]

        assert np.all(threshold_sp >= 0.)
        # for now, enforce threshold as a float
        assert type(threshold_sp) in (float, int)
        try:
            self.sp_crit = float(threshold_sp)
        except TypeError:
            try:
                self.sp_crit = self.grid.at_node[threshold_sp]
            except TypeError:  # was an array
                self.sp_crit = threshold_sp
                assert self.sp_crit.size == self.grid.number_of_nodes
        if np.any(threshold_sp != 0.):
            self.set_threshold = True
            # ^flag for sed_flux_dep_incision to see if the threshold was
            # manually set.
        else:
            self.set_threshold = False
        try:
            self.tstep = kwds['dt']
        except KeyError:
            self.tstep = None
            # retained for back compatibility; undocumented functionality
        if type(use_W) is bool:  # again for back-compatibility
            self.use_W = use_W
            self._W = None
        elif use_W is None:
            self.use_W = False
            self._W = None
        else:
            self.use_W = True
            try:
                self._W = self.grid.at_node[use_W]
            except (FieldError, TypeError):
                assert use_W.size == self._grid.number_of_nodes
                self._W = use_W
        if type(use_Q) is bool:
            self.use_Q = use_Q
            self._Q = None
        elif use_Q is None:
            self.use_Q = False
            self._Q = None
        else:
            self.use_Q = True
            try:
                self._Q = self.grid.at_node[use_Q]
            except (FieldError, TypeError):
                assert use_Q.size == self._grid.number_of_nodes
                self._Q = use_Q
        self._type = sp_type
        if sp_type is 'set_mn':
            assert (float(m_sp) >= 0.) and (float(n_sp) >= 0.), \
                "m and n must be positive"
            self._m = float(m_sp)
            self._n = float(n_sp)
            assert ((a_sp is None) and (b_sp is None) and (c_sp is None)), (
                "If sp_type is 'set_mn', do not pass values for a, b, or c!")
        else:
            assert sp_type in ('Total', 'Unit', 'Shear_stress'), (
                "sp_type not recognised. It must be 'set_mn', 'Total', " +
                "'Unit', or 'Shear_stress'.")
            assert (m_sp == 0.5 and n_sp == 1.), \
                "Do not set m and n if sp_type is not 'set_mn'!"
            assert float(a_sp) >= 0., "a must be positive"
            self._a = float(a_sp)
            if b_sp is not None:
                assert float(b_sp) >= 0., "b must be positive"
                self._b = float(b_sp)
            else:
                assert self.use_W, "b was not set"
                self._b = 0.
            if c_sp is not None:
                assert float(c_sp) >= 0., "c must be positive"
                self._c = float(c_sp)
            else:
                assert self.use_Q, "c was not set"
                self._c = 1.
            if self._type == 'Total':
                self._n = self._a
                self._m = self._a*self._c  # ==_a if use_Q
            elif self._type == 'Unit':
                self._n = self._a
                self._m = self._a*self._c*(1.-self._b)
                # ^ ==_a iff use_Q&use_W etc
            elif self._type == 'Shear_stress':
                self._m = 2.*self._a*self._c*(1.-self._b)/3.
                self._n = 2.*self._a/3.
            else:
                raise MissingKeyError('Not enough information was provided ' +
                                      'on the exponents to use!')
        # m and n will always be set, but care needs to be taken to include Q
        # and W directly if appropriate

        self.stream_power_erosion = grid.zeros(centering='node')
        self.alpha = self.grid.zeros('node')
        self.alpha_divided = self.grid.zeros('node')

    def erode(self, grid, dt, node_elevs='topographic__elevation',
              node_drainage_areas='drainage_area',
              flow_receiver='flow__receiver_node',
              node_order_upstream='flow__upstream_node_order',
              slopes_at_nodes='topographic__steepest_slope',
              link_node_mapping='flow__link_to_receiver_node',
              link_slopes=None, slopes_from_elevs=None,
              W_if_used=None, Q_if_used=None, K_if_used=None,
              flooded_nodes=None):
        """
        .. note:: deprecated
            This run method is now DEPRECATED. Use the fully standardized
            method :func:`run_one_step` instead.

        A simple, explicit implementation of a stream power algorithm.

        Parameters
        ----------
        grid : RasterModelGrid
            A grid.
        dt : float
            Component time step.

        node_elevs : str or ndarray, optional
            Elevations on the grid, either a field string or nnodes-long array.

        node_drainage_areas: str or ndarray, optional
            Tells the component where to look for the drainage area values.
            Change to another string to override which grid field the
            component looks at, or pass a nnodes-long array of drainage
            areas values directly instead.

        flow_receiver, node_order_upstream : str or ndarray, optional
            The downstream node to which each node flows and the ordering of
            the nodes in the network starting at the outlet, respectively,
            are both necessary as inputs to allow stability testing.

            If you already have slopes defined at nodes on the grid, pass them
            to the component with *slopes_at_nodes*. The same syntax is
            expected: string gives a name in the grid fields, an array gives
            values direct.

            Alternatively, set *link_slopes* (and *link_node_mapping*) if this
            data
            is only available at links. 'topographic__derivative_of_elevation'
            is the default field name for link slopes. Override this name by
            setting the variable as the appropriate string, or override use of
            grid fields altogether by passing an array. *link_node_mapping*
            controls how the component maps these link values onto the arrays.
            We assume there is always a 1:1 mapping (pass the values already
            projected onto the nodes using slopes_at_nodes if not). Other
            components, e.g., flow_routing.route_flow_dn, may provide the
            necessary outputs to make the mapping easier: e.g., just pass
            'flow__link_to_receiver_node' from that module (the default name). If
            the component cannot find an existing mapping through this
            parameter, it will derive one on the fly, at considerable cost of
            speed (see on-screen reports).

        slopes_from_elevs : str, optional
            Allows the module to create gradients internally
            from elevations rather than have them provided. Set to True to
            force the component to look for the data in the location specified
            by node_elevs. Using this option is
            considerably slower than any of the alternatives, as it also has to
            calculate the link_node_mapping from stratch each time.

            In both these cases, at present the mapping is to use the maximum
            slope of *any* link attached to the node as the representative
            node slope. This is primarily for speed, but may be a good idea
            to modify later.

        W_if_used, Q_if_used : str or ndarray, optional
            Must be provided if you set *use_W* and *use_Q* respectively in
            the component initialization. They can be either field names or
            nnodes arrays as in the other cases.

            If you are routing across flooded depressions in your flow routing
            scheme, be sure to set *flooded_nodes* with a boolean array or
            array of IDs to ensure erosion cannot occur in the lake. Erosion
            is always zero if the gradient is adverse, but can still procede as
            usual on the entry into the depression unless *flooded_nodes* is
            set.

            NB: If you want spatially or temporally variable runoff, pass the
            runoff values at each pixel to the flow router, then pass
            discharges at each node using *Q_if_used* to this component.

        Returns
        -------
        tuple
            Tuple of (*grid*, *modified_elevs*, *stream_power_erosion*);
            modifies grid elevation fields to reflect updates. Note the value
            stream_power_erosion is not an excess stream power; any specified
            erosion threshold is not incorporated into it.
        """
        upstream_order_IDs = self._grid['node']['flow__upstream_node_order']
        defined_flow_receivers = np.not_equal(self._grid['node'][
            'flow__link_to_receiver_node'], UNDEFINED_INDEX)
        flow_link_lengths = self._grid._length_of_link_with_diagonals[
            self._grid['node']['flow__link_to_receiver_node'][
                defined_flow_receivers]]
        active_nodes = np.where(grid.status_at_node != CLOSED_BOUNDARY)[0]
        flow_receivers = self.grid['node']['flow__receiver_node']

        if W_if_used is not None:
            assert self.use_W, ("Widths were provided, but you didn't set " +
                                "the use_W flag in your input file! " +
                                "Aborting...")
            assert self._W is None, ("Do not pass W to the run method " +
                                     "if you also set them at initialization!")

        if Q_if_used is not None:
            assert self.use_Q, ("Discharges were provided, but you didn't " +
                                "set the use_Q flag in your input file! " +
                                "Aborting...")
            assert self._Q is None, ("Do not pass Q to the run method " +
                                     "if you also set them at initialization!")

        if K_if_used is not None:
            assert self.use_K, ("An array of erodabilities was provided, " +
                                "but you didn't set K_sp to 'array' in your " +
                                "input file! Aborting...")
            try:
                self._K_unit_time = grid.at_node[K_if_used]  # [active_nodes]
            except TypeError:
                self._K_unit_time = K_if_used  # [active_nodes]

        if type(node_elevs) is str:
            node_z = grid.at_node[node_elevs]
        else:
            node_z = node_elevs

        if type(node_drainage_areas) is str:
            node_A = grid.at_node[node_drainage_areas]
        else:
            node_A = node_drainage_areas

        if type(node_order_upstream) is str:
            node_order_upstream = grid.at_node[node_order_upstream]

        # Disable incision in flooded nodes, as appropriate
        if flooded_nodes is not None:
            self._K_unit_time[flooded_nodes] = 0.

        # Operate the main function:
        if self.use_W is False and self.use_Q is False:  # normal case
            self.alpha[defined_flow_receivers] = self._K_unit_time[
                defined_flow_receivers]*dt*node_A[
                    defined_flow_receivers]**self._m / flow_link_lengths
            # Handle flooded nodes, if any (no erosion there)
            if flooded_nodes is not None:
                self.alpha[flooded_nodes] = 0.
            else:
                reversed_flow = node_z < node_z[flow_receivers]
                # this check necessary if flow has been routed across
                # depressions
                self.alpha[reversed_flow] = 0.
            self.alpha_divided[defined_flow_receivers] = (
                self.alpha[defined_flow_receivers] /
                flow_link_lengths**(self._n - 1.))
            threshdt = self.sp_crit * dt
            if type(threshdt) is float:
                erode_with_link_alpha_fixthresh(upstream_order_IDs,
                                                flow_receivers,
                                                threshdt, self.alpha_divided,
                                                self._n, node_z)
            else:
                erode_with_link_alpha_varthresh(upstream_order_IDs,
                                                flow_receivers,
                                                threshdt, self.alpha_divided,
                                                self._n, node_z)
            # stream_power_active_nodes = (self._K_unit_time * dt *
            #                              node_A[active_nodes]**self._m *
            #                              self.slopes[active_nodes]**self._n)
        elif self.use_W:
            if self._W is None:
                try:
                    W = grid.at_node[W_if_used]
                except TypeError:
                    W = W_if_used
            else:
                W = self._W
            if self.use_Q:  # use both Q and W direct
                if self._Q is None:
                    try:
                        Q_direct = grid.at_node[Q_if_used]
                    except TypeError:
                        Q_direct = Q_if_used
                else:
                    Q_direct = self._Q
                self.alpha[defined_flow_receivers] = (
                    self._K_unit_time[defined_flow_receivers]*dt *
                    Q_direct[defined_flow_receivers]**self._m /
                    W[defined_flow_receivers] / flow_link_lengths)
                # Handle flooded nodes, if any (no erosion there)
                if flooded_nodes is not None:
                    self.alpha[flooded_nodes] = 0.
                else:
                    reversed_flow = node_z < node_z[flow_receivers]
                    # this check necessary if flow has been routed across
                    # depressions
                    self.alpha[reversed_flow] = 0.
                self.alpha_divided[defined_flow_receivers] = (
                    self.alpha[defined_flow_receivers] /
                    flow_link_lengths**(self._n - 1.))
                threshdt = self.sp_crit * dt
                if type(threshdt) is float:
                    erode_with_link_alpha_fixthresh(
                        upstream_order_IDs, flow_receivers, threshdt,
                        self.alpha_divided, self._n, node_z)
                else:
                    erode_with_link_alpha_varthresh(
                        upstream_order_IDs, flow_receivers, threshdt,
                        self.alpha_divided, self._n, node_z)
                # stream_power_active_nodes = (
                #     self._K_unit_time*dt*Q_direct[active_nodes]**self._m *
                #     self.slopes[active_nodes]**self._n / W[active_nodes])
            else:  # just W to be used
                self.alpha[defined_flow_receivers] = (
                    self._K_unit_time[defined_flow_receivers]*dt *
                    node_A[defined_flow_receivers]**self._m /
                    W[defined_flow_receivers] / flow_link_lengths)
                # Handle flooded nodes, if any (no erosion there)
                if flooded_nodes is not None:
                    self.alpha[flooded_nodes] = 0.
                else:
                    reversed_flow = node_z < node_z[flow_receivers]
                    # this check necessary if flow has been routed across
                    # depressions
                    self.alpha[reversed_flow] = 0.
                self.alpha_divided[defined_flow_receivers] = (
                    self.alpha[defined_flow_receivers] /
                    flow_link_lengths**(self._n - 1.))
                threshdt = self.sp_crit * dt
                if type(threshdt) is float:
                    erode_with_link_alpha_fixthresh(
                        upstream_order_IDs, flow_receivers, threshdt,
                        self.alpha_divided, self._n, node_z)
                else:
                    erode_with_link_alpha_varthresh(
                        upstream_order_IDs, flow_receivers, threshdt,
                        self.alpha_divided, self._n, node_z)
                # stream_power_active_nodes = (
                #     self._K_unit_time*dt*node_A[active_nodes]**self._m *
                #     self.slopes[active_nodes]**self._n / W[active_nodes])
        else:  # just use_Q
            if self._Q is None:
                try:
                    Q_direct = grid.at_node[Q_if_used]
                except TypeError:
                    assert type(Q_if_used) in (np.ndarray, list)
                    Q_direct = Q_if_used
            else:
                Q_direct = self._Q
                self.alpha[defined_flow_receivers] = (
                    self._K_unit_time[defined_flow_receivers]*dt *
                    Q_direct[defined_flow_receivers]**self._m /
                    flow_link_lengths)
                # Handle flooded nodes, if any (no erosion there)
                if flooded_nodes is not None:
                    self.alpha[flooded_nodes] = 0.
                else:
                    reversed_flow = node_z < node_z[flow_receivers]
                    # this check necessary if flow has been routed across
                    # depressions
                    self.alpha[reversed_flow] = 0.
                self.alpha_divided[defined_flow_receivers] = (
                    self.alpha[defined_flow_receivers] /
                    flow_link_lengths**(self._n - 1.))
                threshdt = self.sp_crit * dt
                if type(threshdt) is float:
                    erode_with_link_alpha_fixthresh(
                        upstream_order_IDs, flow_receivers, threshdt,
                        self.alpha_divided, self._n, node_z)
                else:
                    erode_with_link_alpha_varthresh(
                        upstream_order_IDs, flow_receivers, threshdt,
                        self.alpha_divided, self._n, node_z)
            # stream_power_active_nodes = (
            #     self._K_unit_time*dt*Q_direct[active_nodes]**self._m *
            #     self.slopes[active_nodes]**self._n)

        # # Note that we save "stream_power_erosion" incorporating both K and a.
        # # Most definitions would need this value /K then **(1/a) to give actual
        # # stream power (unit, total, whatever), and it does not yet include the
        # # threshold
        # self.stream_power_erosion[active_nodes] = stream_power_active_nodes
        # grid.at_node['stream_power_erosion'][:] = self.stream_power_erosion
        # erosion_increment = (self.stream_power_erosion - self.sp_crit).clip(0.)
        # 
        # # this prevents any node from incising below any node downstream of it
        # # we have to go in upstream order in case our rate is so big we impinge
        # # on baselevels > 1 node away
        # 
        # elev_dstr = node_z[flow_receiver]
        # # ^we substract erosion_increment[flow_receiver] in the loop, as it
        # # can update
        # 
        # method = 'cython'
        # if method == 'cython':
        #     from .cfuncs import erode_avoiding_pits
        # 
        #     erode_avoiding_pits(node_order_upstream, flow_receiver, node_z,
        #                         erosion_increment)
        # else:
        #     for i in node_order_upstream:
        #         elev_this_node_before = node_z[i]
        #         elev_this_node_after = (elev_this_node_before -
        #                                 erosion_increment[i])
        #         elev_dstr_node_after = (elev_dstr[i] -
        #                                 erosion_increment[flow_receiver[i]])
        #         if elev_this_node_after < elev_dstr_node_after:
        #             erosion_increment[i] = (elev_this_node_before -
        #                                     elev_dstr_node_after)*0.999999
        #         # ^we add a tiny elevation excess to prevent the module from
        #         # ever totally severing its own flow paths
        # # clip the erosion increments one more time to remove regatives
        # # introduced by any pit filling algorithms or the above procedure:
        # node_z -= erosion_increment.clip(0.)
        # 
        # self._grid = grid

        return grid, node_z, self.stream_power_erosion

    def run_one_step(self, dt, flooded_nodes=None, **kwds):
        """
        A simple, explicit implementation of a stream power algorithm.

        This component now looks exclusively for the field
        'topographic__steepest_slope' at each node to determine the local
        slope (previoiusly it was possible to map values from links explicitly
        within the component, but this functionality is now deprecated).

        If you are routing across flooded depressions in your flow routing
        scheme, be sure to set *flooded_nodes* with a boolean array or array
        of IDs to ensure erosion cannot occur in the lake. Erosion
        is always zero if the gradient is adverse, but can still procede as
        usual on the entry into the depression unless *flooded_nodes* is set.

        Parameters
        ----------
        dt : float
            Time-step size
        flooded_nodes : ndarray of int (optional)
            IDs of nodes that are flooded and should have no erosion. If not
            provided but flow has still been routed across depressions, erosion
            may still occur beneath the apparent water level (though will
            always still be positive).
        """
        self.erode(grid=self._grid, dt=dt, flooded_nodes=flooded_nodes)



#! /usr/bin/env python
"""Write structured grids to NetCDF files.

Write netcdf
++++++++++++

.. autosummary::
    :toctree: generated/

    ~landlab.io.netcdf.write.write_netcdf
"""


import os
import warnings
import six

import numpy as np

try:
    import netCDF4 as nc4
except ImportError:
    warnings.warn('Unable to import netCDF4.', ImportWarning)

from scipy.io import netcdf as nc


from landlab.io.netcdf._constants import (_AXIS_DIMENSION_NAMES,
                                          _AXIS_COORDINATE_NAMES,
                                          _NP_TO_NC_TYPE)


def _set_netcdf_attributes(root, attrs):
    """Set attributes of a netcdf file.

    Set attributes of the netCDF Database object, *root*. Attributes are
    given as key/value pairs from *attrs*.

    Parameters
    ----------
    root : netcdf_file
        A NetCDF file.
    attrs : dict
        Attributes as key-value pairs.
    """
    for (key, val) in attrs.items():
        setattr(root, key, val)


def _get_dimension_names(shape):
    """Get dimension names.

    Parameters
    ----------
    shape : tuple of int
        Shape of a structured grid.

    Returns
    -------
    tuple of str
        Dimension names for the NetCDF file.

    Examples
    --------
    >>> from landlab.io.netcdf.write import _get_dimension_names
    >>> _get_dimension_names((4, ))
    ['ni']
    >>> _get_dimension_names((4, 5))
    ['nj', 'ni']
    >>> _get_dimension_names((4, 5, 6))
    ['nk', 'nj', 'ni']
    """
    names = _AXIS_DIMENSION_NAMES[- 1: - (len(shape) + 1): - 1]
    return names[::-1]


def _get_dimension_sizes(shape):
    """Get dimension sizes.

    Parameters
    ----------
    shape : tuple of int
        Shape of a structured grid.

    Returns
    -------
    dict
        Dimension sizes.

    Examples
    --------
    >>> from landlab.io.netcdf.write import _get_dimension_sizes
    >>> _get_dimension_sizes((4, ))
    {'ni': 4}
    >>> sizes = _get_dimension_sizes((4, 5))
    >>> sizes['ni'], sizes['nj']
    (5, 4)
    >>> sizes = _get_dimension_sizes((4, 5, 6))
    >>> sizes['ni'], sizes['nj'], sizes['nk']
    (6, 5, 4)
    """
    names = _AXIS_DIMENSION_NAMES[- 1: - (len(shape) + 1): - 1]

    sizes = dict()
    for (axis, name) in enumerate(names):
        sizes[name] = shape[- (axis + 1)]

    return sizes


def _get_axes_names(shape):
    """Get names of the axes.

    Parameters
    ----------
    shape : tuple of int
        Shape of a structured grid.

    Returns
    -------
    tuple of str
        Names of the axes for the NetCDF file.

    Examples
    --------
    >>> from landlab.io.netcdf.write import _get_axes_names
    >>> _get_axes_names((2, ))
    ['x']
    >>> _get_axes_names((2, 3))
    ['y', 'x']
    >>> _get_axes_names((2, 3, 4))
    ['z', 'y', 'x']
    """
    names = _AXIS_COORDINATE_NAMES[- 1: - (len(shape) + 1): - 1]
    return names[::-1]


def _get_cell_bounds(shape, spacing=(1., 1.), origin=(0., 0.)):
    """Get bounds arrays for square cells.

    Parameters
    ----------
    shape : tuple of int
        Shape of the grid in cell corners.
    spacing : tuple of float
        Height and width of cells.
    origin : tuple of float
        Coordinates of lower-left corner of lower-left cell.

    Returns
    -------
    (y, x) : tuple of ndarray
        Tuple of the *y* and *x* coordinates of each cell corner (ordered
        counter-clockwise starting from lower-right. The shape of the returned
        arrays will be *(rows, cols, 4)*.

    Examples
    --------
    >>> from landlab.io.netcdf.write import _get_cell_bounds
    >>> bounds = _get_cell_bounds((3, 4))
    >>> bounds['y_bnds'] # doctest: +NORMALIZE_WHITESPACE
    array([[[ 0.,  1.,  1.,  0.], [ 0.,  1.,  1.,  0.], [ 0.,  1.,  1.,  0.]],
           [[ 1.,  2.,  2.,  1.], [ 1.,  2.,  2.,  1.], [ 1.,  2.,  2.,  1.]]])
    >>> bounds['x_bnds'] # doctest: +NORMALIZE_WHITESPACE
    array([[[ 1.,  1.,  0.,  0.], [ 2.,  2.,  1.,  1.], [ 3.,  3.,  2.,  2.]],
           [[ 1.,  1.,  0.,  0.], [ 2.,  2.,  1.,  1.], [ 3.,  3.,  2.,  2.]]])
    """
    rows = np.arange(shape[0]) * spacing[0] + origin[0]
    cols = np.arange(shape[1]) * spacing[1] + origin[1]

    corner_y, corner_x = np.meshgrid(rows, cols, indexing='ij')

    y_bnds = np.vstack((corner_y[:-1, 1:].flat, corner_y[1:, 1:].flat,
                        corner_y[1:, :-1].flat, corner_y[:-1, :-1].flat)).T
    x_bnds = np.vstack((corner_x[:-1, 1:].flat, corner_x[1:, 1:].flat,
                        corner_x[1:, :-1].flat, corner_x[:-1, :-1].flat)).T

    return {'y_bnds': y_bnds.reshape((shape[0] - 1, shape[1] - 1, 4)),
            'x_bnds': x_bnds.reshape((shape[0] - 1, shape[1] - 1, 4)),
           }


def _set_netcdf_cell_structured_dimensions(root, shape):
    """Set dimensions for a structured grid of cells.

    Parameters
    ----------
    root : netcdf_file
        A NetCDF file.
    shape : tuple of int
        Shape of the cell grid (rows of cells, columns of cells).
    """
    if len(shape) < 1 or len(shape) > 3:
        raise ValueError('grid dimension must be 1, 2, or 3')

    dimensions = _get_dimension_sizes(shape)

    dims = root.dimensions

    if 'nt' not in dims:
        root.createDimension('nt', None)

    for (name, dim_size) in dimensions.items():
        if name not in dims:
            root.createDimension(name, dim_size - 2)

    root.createDimension('nv', 4)


def _set_netcdf_structured_dimensions(root, shape):
    """Set dimensions for a structured grid.

    Add dimensions to *root* for a structured grid of size *shape*. The
    dimension names will be 'ni', 'nj', and 'nk'. 'ni' is the length of the
    fast dimension, followed by 'nj', and then 'nk'.

    For example, a grid with shape (3, 4, 5) will have dimensions ni=5,
    nj=4, and nk=3. Lower dimension grids simply drop the slowest dimension.
    Thus, a grid with shape (3, 4) has dimensions ni=4, and nj=3.

    Parameters
    ----------
    root : netcdf_file
        A NetCDF file.
    shape : tuple of int
        Shape of the grid.
    """
    if len(shape) < 1 or len(shape) > 3:
        raise ValueError('grid dimension must be 1, 2, or 3')

    dimensions = _get_dimension_sizes(shape)

    dims = root.dimensions

    if 'nt' not in dims:
        root.createDimension('nt', None)

    for (name, dim_size) in dimensions.items():
        if name not in dims:
            root.createDimension(name, dim_size)


def _set_netcdf_variables(root, fields, **kwds):
    """Set the field variables.

    First set the variables that define the grid and then the variables at
    the grid nodes and cells.
    """
    names = kwds.pop('names', None)

    _add_spatial_variables(root, fields, **kwds)
    _add_variables_at_points(root, fields, names=names)


def _set_netcdf_cell_variables(root, fields, **kwds):
    """Set the cell field variables.

    First set the variables that define the grid and then the variables at
    the grid nodes and cells.
    """
    names = kwds.pop('names', None)

    _add_cell_spatial_variables(root, fields, **kwds)
    _add_variables_at_cells(root, fields, names=names)


def _add_cell_spatial_variables(root, grid, **kwds):
    """Add the spatial variables that describe the cell grid."""
    long_name = kwds.get('long_name', {})

    cell_grid_shape = [dim - 1 for dim in grid.shape]
    spatial_variable_shape = _get_dimension_names(cell_grid_shape)

    bounds = _get_cell_bounds(cell_grid_shape,
                              spacing=(grid.dy, grid.dx),
                              origin=(grid.dy * .5, grid.dx * .5))

    shape = spatial_variable_shape + ['nv']
    for name, values in bounds.items():
        # var = root.createVariable(name, 'f8', shape)
        # var[:] = values

        try:
            var = root.variables[name]
        except KeyError:
            var = root.createVariable(name, 'f8', shape)

        var[:] = values

        axis = grid.axis_name.index(name[0])

        var.units = grid.axis_units[axis]
        try:
            var.long_name = long_name[name]
        except KeyError:
            var.long_name = grid.axis_name[axis]


def _add_spatial_variables(root, grid, **kwds):
    """Add spatial variables to a NetCDF file.

    Add the variables to *root* that define the structured grid, *grid*.

    Parameters
    ----------
    root : netcdf_file
        A NetCDF file.
    grid : RasterModelGrid
        A structured grid.
    long_name : dict, optional
        Long name for each spatial variable to add. Keys are grid field
        names, values are corresponding long names.
    """
    long_name = kwds.get('long_name', {})

    netcdf_vars = root.variables

    spatial_variable_names = _get_axes_names(grid.shape)
    spatial_variable_shape = _get_dimension_names(grid.shape)

    for (axis, name) in enumerate(spatial_variable_names):
        try:
            var = netcdf_vars[name]
        except KeyError:
            var = root.createVariable(name, 'f8', spatial_variable_shape)

        coords = grid.node_axis_coordinates(axis=axis).view()
        coords.shape = var.shape
        var[:] = coords

        var.units = grid.axis_units[axis]
        try:
            var.long_name = long_name[name]
        except KeyError:
            var.long_name = grid.axis_name[axis]


def _add_variables_at_points(root, fields, names=None):
    if isinstance(names, six.string_types):
        names = [names]
    names = names or fields['node'].keys()

    netcdf_vars = root.variables

    spatial_variable_shape = _get_dimension_names(fields.shape)

    try:
        n_times = len(netcdf_vars['t']) - 1
    except KeyError:
        n_times = 0

    node_fields = fields['node']
    for var_name in names:
        try:
            var = netcdf_vars[var_name]
        except KeyError:
            var = root.createVariable(
                var_name, _NP_TO_NC_TYPE[str(node_fields[var_name].dtype)],
                ['nt'] + spatial_variable_shape)

        if node_fields[var_name].size > 1:
            data = node_fields[var_name].view()
            data.shape = var.shape[1:]
            try:
                var[n_times, :] = data
            except ValueError:
                raise
        else:
            var[n_times] = node_fields[var_name].flat[0]

        var.units = node_fields.units[var_name] or '?'
        var.long_name = var_name


def _add_variables_at_cells(root, fields, names=None):
    if isinstance(names, six.string_types):
        names = [names]
    names = names or fields['cell'].keys()

    netcdf_vars = root.variables

    cell_grid_shape = [dim - 1 for dim in fields.shape]

    spatial_variable_shape = _get_dimension_names(cell_grid_shape)

    try:
        n_times = len(netcdf_vars['t']) - 1
    except KeyError:
        n_times = 0

    cell_fields = fields['cell']
    for var_name in names:
        try:
            var = netcdf_vars[var_name]
        except KeyError:
            var = root.createVariable(
                var_name, _NP_TO_NC_TYPE[str(cell_fields[var_name].dtype)],
                ['nt'] + spatial_variable_shape)

        if cell_fields[var_name].size > 1:
            data = cell_fields[var_name].view()
            data.shape = var.shape[1:]
            try:
                var[n_times, :] = data
            except ValueError:
                raise
        else:
            var[n_times] = cell_fields[var_name].flat[0]

        var.units = cell_fields.units[var_name] or '?'
        var.long_name = var_name


def _add_time_variable(root, time, **kwds):
    """Add a time value to a NetCDF file.

    Append a new time value to the time variable of a NetCDF file. If there
    is not time variable, create one. The time variable is named, ``t``.

    Parameters
    ----------
    root : netcdf_file
        A NetCDF file.
    time : float
        The time.
    units : str, optional
        Time units.
    reference : str, optional
        Reference time.
    """
    units = kwds.get('units', 'days')
    reference = kwds.get('reference', '00:00:00 UTC')

    netcdf_vars = root.variables

    try:
        time_var = netcdf_vars['t']
    except KeyError:
        time_var = root.createVariable('t', 'f8', ('nt', ))
        time_var.units = ' '.join([units, 'since', reference])
        time_var.long_name = 'time'

    n_times = len(time_var)
    if time is not None:
        time_var[n_times] = time
    else:
        time_var[n_times] = n_times


_VALID_NETCDF_FORMATS = set([
    'NETCDF3_CLASSIC',
    'NETCDF3_64BIT',
    'NETCDF4_CLASSIC',
    'NETCDF4',
])


def _guess_at_location(fields, names):
    """Guess where the values should be located."""
    node_fields = set(fields['node'].keys())
    cell_fields = set(fields['cell'].keys())

    if names is None or len(names) == 0:
        if len(fields['node']) > 0:
            at = 'node'
        else:
            at = 'cell'
    else:
        if node_fields.issuperset(names):
            at = 'node'
        elif cell_fields.issuperset(names):
            at = 'cell'
        else:
            at = None
    return at


def write_netcdf(path, fields, attrs=None, append=False,
                 format='NETCDF3_64BIT', names=None, at=None):
    """Write landlab fields to netcdf.

    Write the data and grid information for *fields* to *path* as NetCDF.
    If the *append* keyword argument in True, append the data to an existing
    file, if it exists. Otherwise, clobber an existing files.

    Parameters
    ----------
    path : str
        Path to output file.
    fields : field-like
        Landlab field object that holds a grid and associated values.
    append : boolean, optional
        Append data to an existing file, otherwise clobber the file.
    format : {'NETCDF3_CLASSIC', 'NETCDF3_64BIT', 'NETCDF4_CLASSIC', 'NETCDF4'}
        Format of output netcdf file.
    attrs : dict
        Attributes to add to netcdf file.
    names : iterable of str, optional
        Names of the fields to include in the netcdf file. If not provided,
        write all fields.
    at : {'node', 'cell'}, optional
        The location where values are defined.

    Examples
    --------
    >>> import numpy as np
    >>> from landlab import RasterModelGrid
    >>> from landlab.io.netcdf import write_netcdf

    Create a uniform rectilinear grid with four rows and 3 columns, and add
    some data fields to it.

    >>> rmg = RasterModelGrid(4, 3)
    >>> _ = rmg.add_field('node', 'topographic__elevation', np.arange(12.))
    >>> _ = rmg.add_field('node', 'uplift_rate', 2. * np.arange(12.))

    Create a temporary directory to write the netcdf file into.

    >>> import tempfile, os
    >>> temp_dir = tempfile.mkdtemp()
    >>> os.chdir(temp_dir)

    Write the grid to a netcdf3 file but only include the *uplift_rate*
    data in the file.

    >>> write_netcdf('test.nc', rmg, format='NETCDF3_64BIT',
    ...     names='uplift_rate')

    Read the file back in and check its contents.

    >>> from scipy.io import netcdf
    >>> fp = netcdf.netcdf_file('test.nc', 'r')
    >>> 'uplift_rate' in fp.variables
    True
    >>> 'topographic__elevation' in fp.variables
    False
    >>> fp.variables['uplift_rate'][:].flatten()
    array([  0.,   2.,   4.,   6.,   8.,  10.,  12.,  14.,  16.,  18.,  20.,
            22.])

    >>> _ = rmg.add_field('cell', 'air__temperature', np.arange(2.))
    >>> write_netcdf('test-cell.nc', rmg, format='NETCDF3_64BIT',
    ...     names='air__temperature', at='cell')
    """
    if format not in _VALID_NETCDF_FORMATS:
        raise ValueError('format not understood')
    if at not in (None, 'cell', 'node'):
        raise ValueError('value location not understood')

    if isinstance(names, six.string_types):
        names = (names, )

    at = at or _guess_at_location(fields, names) or 'node'
    names = names or fields[at].keys()

    if not set(fields[at].keys()).issuperset(names):
        raise ValueError('values must be on either cells or nodes, not both')

    attrs = attrs or {}

    if os.path.isfile(path) and append:
        mode = 'a'
    else:
        mode = 'w'

    if format == 'NETCDF3_CLASSIC':
        root = nc.netcdf_file(path, mode, version=1)
    elif format == 'NETCDF3_64BIT':
        root = nc.netcdf_file(path, mode, version=2)
    else:
        root = nc4.Dataset(path, mode, format=format)

    _set_netcdf_attributes(root, attrs)
    if at == 'node':
        _set_netcdf_structured_dimensions(root, fields.shape)
        _set_netcdf_variables(root, fields, names=names)
    else:
        _set_netcdf_cell_structured_dimensions(root, fields.shape)
        _set_netcdf_cell_variables(root, fields, names=names)

    root.close()



import numpy as np
from numpy.testing import assert_array_equal
from nose.tools import with_setup, raises

from landlab import RasterModelGrid


def setup():
    globals()['rmg'] = RasterModelGrid(4, 4)


@with_setup(setup)
def test_unit_grid_all_cells():
    assert_array_equal(rmg.area_of_cell, np.ones(4))


@with_setup(setup)
def test_unit_grid_one_cell():
    assert_array_equal(rmg.area_of_cell[0], 1.)


@with_setup(setup)
def test_unit_grid_last_cell():
    assert_array_equal(rmg.area_of_cell[-1], 1.)


@with_setup(setup)
@raises(IndexError)
def test_out_of_range():
    rmg.area_of_cell[5]


@with_setup(setup)
@raises(ValueError)
def test_is_immutable():
    rmg.area_of_cell[0] = 0.


def test_all_cells_with_spacing():
    rmg = RasterModelGrid(4, 4, 10.)
    assert_array_equal(rmg.area_of_cell, 100 * np.ones(4))



#! /usr/bin/env python
"""Read data from a pickled Landlab grid file into a RasterModelGrid.

Read Landlab native
+++++++++++++++++++

.. autosummary::
    :toctree: generated/

    ~landlab.io.native_landlab.load_grid
    ~landlab.io.native_landlab.save_grid
"""

import os
from six.moves import cPickle
from landlab import ModelGrid


def save_grid(grid, path, clobber=False):
    """Save a grid and fields to a Landlab "native" format.

    This method uses cPickle to save a grid as a cPickle file.
    All fields will be saved, along with the grid.

    The recommended suffix for the save file is '.grid'. This will
    be added to your save if you don't include it.

    Caution: Pickling can be slow, and can produce very large files.
    Caution 2: Future updates to Landlab could potentially render old
    saves unloadable.

    Parameters
    ----------
    grid : object of subclass ModelGrid
        Grid object to save
    path : str
        Path to output file, either without suffix, or '.grid'
    clobber : bool (default False)
        Set to True to allow overwrites of existing files

    Examples
    --------
    >>> from landlab import RasterModelGrid
    >>> from landlab.io.native_landlab import save_grid
    >>> import os
    >>> grid_out = RasterModelGrid(4,5,2.)
    >>> save_grid(grid_out, 'testsavedgrid.grid', clobber=True)
    >>> os.remove('testsavedgrid.grid') #to remove traces of this test
    """
    if os.path.exists(path) and not clobber:
        raise ValueError('file exists')

    # test it's a grid
    assert issubclass(type(grid), ModelGrid)

    (base, ext) = os.path.splitext(path)
    if ext != '.grid':
        ext = ext + '.grid'
    path = base + ext

    with open(path, 'wb') as file_like:
        cPickle.dump(grid, file_like)


def load_grid(path):
    """Load a grid and its fields from a Landlab "native" format.

    This method uses cPickle to load a saved grid.
    It assumes you saved using vmg.save() or save_grid, i.e., that the
    pickle file is a .grid file.

    Caution: Pickling can be slow, and can produce very large files.
    Caution 2: Future updates to Landlab could potentially render old
    saves unloadable.

    Parameters
    ----------
    path : str
        Path to output file, either without suffix, or '.grid'

    Examples
    --------
    >>> from landlab import VoronoiDelaunayGrid
    >>> from landlab.io.native_landlab import load_grid, save_grid
    >>> import numpy as np
    >>> import os
    >>> x = np.random.rand(20)
    >>> y = np.random.rand(20)
    >>> grid_out = VoronoiDelaunayGrid(x, y)
    >>> save_grid(grid_out, 'testsavedgrid.grid', clobber=True)
    >>> grid_in = load_grid('testsavedgrid.grid')
    >>> os.remove('testsavedgrid.grid') #to remove traces of this test
    """
    (base, ext) = os.path.splitext(path)
    if ext != '.grid':
        ext = ext + '.grid'
    path = base + ext
    with open(path, 'rb') as file_like:
        loaded_grid = cPickle.load(file_like)
    assert issubclass(type(loaded_grid), ModelGrid)
    return loaded_grid



#! /usr/bin/env python
"""Calculate steepest descent on a raster grid.

Steepest descent functions for raster grids
+++++++++++++++++++++++

.. autosummary::
    :toctree: generated/

    ~landlab.grid.raster_steepest_descent._calc_steepest_descent_across_adjacent_cells
    ~landlab.grid.raster_steepest_descent._calc_steepest_descent_across_cell_corners
    ~landlab.grid.raster_steepest_descent._calc_steepest_descent_across_cell_faces

"""
import numpy as np

from landlab.core.utils import make_optional_arg_into_id_array
from landlab.utils.decorators import deprecated
from landlab.grid.raster_gradients import (
    calc_grad_across_cell_faces,
    calc_grad_across_cell_corners,
    calc_grad_along_node_links)


_VALID_ROUTING_METHODS = set(['d8', 'd4'])


def _assert_valid_routing_method(method):
    """Check if name is a valid routing method.

    Parameters
    ----------
    method : str
        Name of a routing method.

    Raises
    ------
    ValueError
        If the method name is invalid.
    """
    if method not in _VALID_ROUTING_METHODS:
        raise ValueError(
            '%s: routing method not understood. should be one of %s' %
            (method, ', '.join(_VALID_ROUTING_METHODS)))


def _calc_steepest_descent_across_adjacent_cells(grid, node_values, *args,
                                                **kwds):
    """Get steepest gradient to neighbor and diagonal cells.

    Calculate the steepest downward gradients in *node_values*, given at every
    node in the grid, relative to the nodes centered at *cell_ids*. Note that
    upward gradients are reported as positive, so this method returns negative
    numbers.

    If *cell_ids* is not provided, calculate the maximum gradient for all
    cells in the grid.

    The default is to only consider neighbor cells to the north, south, east,
    and west. To also consider gradients to diagonal nodes, set the *method*
    keyword to *d8* (the default is *d4*).

    Use the *out* keyword if you have an array that you want to put the result
    into. If not given, create a new array.

    Use the *return_node* keyword to also the node id of the node in the
    direction of the maximum gradient.

    Parameters
    ----------
    grid : RasterModelGrid
        Input grid.
    node_values : array_like
        Values to take gradient of.
    cell_ids : array_like, optional
        IDs of grid cells to measure gradients.
    return_node: boolean, optional
        Return node IDs of the node that has the steepest descent.
    method : {'d4', 'd8'}
        How to calculate the steepest descent.
    out : ndarray, optional
        Alternative output array in which to place the result.  Must
        be of the same shape and buffer length as the expected output.

    Returns
    -------
    ndarray :
        Calculated gradients to lowest adjacent node.

    Examples
    --------
    Create a rectilinear grid that is 3 nodes by 3 nodes and so has one cell
    centered around node 4.

    >>> from landlab import RasterModelGrid
    >>> grid = RasterModelGrid(3, 3)
    >>> values_at_nodes = np.array([-3., -1., 0., 0., 1., 0., 0., 0., 0.])

    Calculate gradients to cell diagonals and choose the gradient to the
    lowest node.

    >>> from math import sqrt
    >>> grid._calc_steepest_descent_across_adjacent_cells(values_at_nodes,
    ...     method='d4')
    masked_array(data = [-2.],
                 mask = False,
           fill_value = 1e+20)
    <BLANKLINE>
    >>> grid._calc_steepest_descent_across_adjacent_cells(values_at_nodes,
    ...     method='d8') * sqrt(2.)
    masked_array(data = [-4.],
                 mask = False,
           fill_value = 1e+20)
    <BLANKLINE>

    With the 'd4' method, the steepest gradient is to the bottom node (id = 1).

    >>> (_, ind) = grid._calc_steepest_descent_across_adjacent_cells(
    ...                values_at_nodes, return_node=True)
    >>> ind
    array([1])

    With the 'd8' method, the steepest gradient is to the lower-left node
    (id = 0).

    >>> (_, ind) = grid._calc_steepest_descent_across_adjacent_cells(
    ...                values_at_nodes, return_node=True, method='d8')
    >>> ind
    array([0])

    >>> from landlab import RasterModelGrid
    >>> grid = RasterModelGrid(4, 4)
    >>> node_values = grid.zeros()
    >>> node_values[1] = -1
    >>> grid._calc_steepest_descent_across_adjacent_cells(node_values, 0)
    masked_array(data = [-1.],
                 mask = False,
           fill_value = 1e+20)
    <BLANKLINE>

    Get both the maximum gradient and the node to which the gradient is
    measured.

    >>> grid._calc_steepest_descent_across_adjacent_cells(
    ...     node_values, 0, return_node=True)
    (array([-1.]), array([1]))

    Use method to choose which neighbors to consider.

    >>> node_values[0] = -10.
    >>> node_values[1] = -1.
    >>> grid._calc_steepest_descent_across_adjacent_cells(
    ...     node_values, 0, method='d4', return_node=True)
    (array([-1.]), array([1]))
    >>> grid._calc_steepest_descent_across_adjacent_cells(
    ...     node_values, 0, method='d8', return_node=True)
    (array([-7.07106781]), array([0]))
    """
    method = kwds.pop('method', 'd4')
    _assert_valid_routing_method(method)

    if method == 'd4':
        return _calc_steepest_descent_across_cell_faces(
            grid, node_values, *args, **kwds)
    elif method == 'd8':
        neighbor_grads = _calc_steepest_descent_across_cell_faces(
            grid, node_values, *args, **kwds)
        diagonal_grads = _calc_steepest_descent_across_cell_corners(
            grid, node_values, *args, **kwds)

        return_node = kwds.pop('return_node', False)

        if not return_node:
            return np.choose(neighbor_grads <= diagonal_grads,
                             (diagonal_grads, neighbor_grads), **kwds)
        else:
            min_grads = np.choose(neighbor_grads[0] <= diagonal_grads[0],
                                  (diagonal_grads[0], neighbor_grads[0]),
                                  **kwds)
            node_ids = np.choose(neighbor_grads[0] <= diagonal_grads[0],
                                 (diagonal_grads[1], neighbor_grads[1]),
                                 **kwds)
            return (min_grads, node_ids)


def _calc_steepest_descent_across_cell_corners(grid, node_values, *args,
                                              **kwds):
    """Get steepest gradient to the diagonals of a cell.

    Calculate the gradients in *node_values* measure to the diagonals of cells
    IDs, *cell_ids*. Slopes upward from the cell are reported as positive.
    If *cell_ids* is not given, calculate gradients for all cells.

    Use the *return_node* keyword to return a tuple, with the first element
    being the gradients and the second the node id of the node in the direction
    of the minimum gradient, i.e., the steepest descent. Note the gradient
    value returned is probably thus negative.

    Parameters
    ----------
    grid : RasterModelGrid
        Input grid.
    node_values : array_like
        Values to take gradient of.
    cell_ids : array_like, optional
        IDs of grid cells to measure gradients.
    return_node: boolean, optional
        If `True`, return node IDs of the node that has the steepest descent.
    out : ndarray, optional
        Alternative output array in which to place the result.  Must
        be of the same shape and buffer length as the expected output.

    Returns
    -------
    ndarray :
        Calculated gradients to lowest node across cell faces.

    Examples
    --------
    Create a rectilinear grid that is 3 nodes by 3 nodes and so has one cell
    centered around node 4.

    >>> from landlab import RasterModelGrid
    >>> grid = RasterModelGrid(3, 3)
    >>> values_at_nodes = np.arange(9.)

    Calculate gradients to cell diagonals and choose the gradient to the
    lowest node.

    >>> from math import sqrt
    >>> grid._calc_steepest_descent_across_cell_corners(
    ...     values_at_nodes) * sqrt(2.)
    array([-4.])

    The steepest gradient is to node with id 0.

    >>> (_, ind) = grid._calc_steepest_descent_across_cell_corners(
    ...                values_at_nodes, return_node=True)
    >>> ind
    array([0])

    >>> grid = RasterModelGrid(3, 3)
    >>> node_values = grid.zeros()
    >>> node_values[0] = -1
    >>> grid._calc_steepest_descent_across_cell_corners(node_values, 0)
    array([-0.70710678])

    Get both the maximum gradient and the node to which the gradient is
    measured.

    >>> grid._calc_steepest_descent_across_cell_corners(node_values, 0,
    ...     return_node=True)
    (array([-0.70710678]), array([0]))
    """
    return_node = kwds.pop('return_node', False)

    cell_ids = make_optional_arg_into_id_array(grid.number_of_cells, *args)

    grads = calc_grad_across_cell_corners(grid, node_values, cell_ids)

    if return_node:
        ind = np.argmin(grads, axis=1)
        node_ids = grid.diagonal_cells[grid.node_at_cell[cell_ids], ind]
        if 'out' not in kwds:
            out = np.empty(len(cell_ids), dtype=grads.dtype)
        out[:] = grads[range(len(cell_ids)), ind]
        return (out, node_ids)
    else:
        return grads.min(axis=1, **kwds)

@deprecated(use='FlowRouter', version=1.0)
def _calc_steepest_descent_across_cell_faces(grid, node_values, *args, **kwds):
    """Get steepest gradient across the faces of a cell.

    This method calculates the gradients in *node_values* across all four
    faces of the cell or cells with ID *cell_ids*. Slopes upward from the
    cell are reported as positive. If *cell_ids* is not given, calculate
    gradients for all cells.

    Use the *return_node* keyword to return a tuple, with the first element
    being the gradients and the second the node id of the node in the direction
    of the minimum gradient, i.e., the steepest descent. Note the gradient
    value returned is probably thus negative.

    Parameters
    ----------
    grid : RasterModelGrid
        Input grid.
    node_values : array_like
        Values to take gradient of.
    cell_ids : array_like, optional
        IDs of grid cells to measure gradients.
    return_node: boolean, optional
        Return node IDs of the node that has the steepest descent.
    out : ndarray, optional
        Alternative output array in which to place the result.  Must
        be of the same shape and buffer length as the expected output.

    Returns
    -------
    ndarray :
        Calculated gradients to lowest node across cell faces.

    Convention: gradients positive UP

    Examples
    --------
    Create a rectilinear grid that is 3 nodes by 3 nodes and so has one cell
    centered around node 4.

    >>> from landlab import RasterModelGrid
    >>> grid = RasterModelGrid(3, 3)
    >>> values_at_nodes = np.arange(9.)

    Calculate gradients across each cell face and choose the gradient to the
    lowest node.

    >>> grid._calc_steepest_descent_across_cell_faces(values_at_nodes)
    masked_array(data = [-3.],
                 mask = False,
           fill_value = 1e+20)
    <BLANKLINE>

    The steepest gradient is to node with id 1.

    >>> (_, ind) = grid._calc_steepest_descent_across_cell_faces(
    ...     values_at_nodes, return_node=True)
    >>> ind
    array([1])

    >>> grid = RasterModelGrid(3, 3)
    >>> node_values = grid.zeros()
    >>> node_values[1] = -1
    >>> grid._calc_steepest_descent_across_cell_faces(node_values, 0)
    masked_array(data = [-1.],
                 mask = False,
           fill_value = 1e+20)
    <BLANKLINE>

    Get both the maximum gradient and the node to which the gradient is
    measured.

    >>> grid._calc_steepest_descent_across_cell_faces(node_values, 0,
    ...     return_node=True)
    (array([-1.]), array([1]))
    """
    return_node = kwds.pop('return_node', False)

    cell_ids = make_optional_arg_into_id_array(grid.number_of_cells, *args)

    grads = calc_grad_across_cell_faces(grid, node_values, cell_ids)

    if return_node:
        ind = np.argmin(grads, axis=1)
        node_ids = grid.active_neighbors_at_node()[grid.node_at_cell[
            cell_ids], ind]
        # node_ids = grid.neighbor_nodes[grid.node_at_cell[cell_ids], ind]
        if 'out' not in kwds:
            out = np.empty(len(cell_ids), dtype=grads.dtype)
        out[:] = grads[range(len(cell_ids)), ind]
        return (out, node_ids)
        # return (out, 3 - ind)
    else:
        return grads.min(axis=1, **kwds)



import numpy as np
from numpy.testing import assert_array_equal
from nose.tools import assert_equal, assert_raises, assert_false, assert_true
try:
    from nose.tools import assert_is, assert_is_instance
except ImportError:
    from landlab.testing.tools import assert_is, assert_is_instance
import landlab.utils.structured_grid as sgrid
from landlab.utils.structured_grid import BAD_INDEX_VALUE


def test_default():
    neighbors = sgrid.neighbor_node_array((2, 3))

    X = sgrid.BAD_INDEX_VALUE
    assert_array_equal(neighbors,
                       np.array([[1, 3, X, X],
                                 [2, 4, 0, X],
                                 [X, 5, 1, X],
                                 [4, X, X, 0],
                                 [5, X, 3, 1],
                                 [X, X, 4, 2]]).T)

    assert_true(neighbors.flags['C_CONTIGUOUS'])
    assert_true(neighbors.base is None)


def test_set_out_of_bounds():
    neighbors = sgrid.neighbor_node_array((2, 3), inactive=-1)
    assert_array_equal(neighbors,
                       np.array([[1,  3, -1, -1],
                                 [2,  4,  0, -1],
                                 [-1,  5,  1, -1],
                                 [4, -1, -1,  0],
                                 [5, -1,  3,  1],
                                 [-1, -1,  4,  2]]).T)


def test_boundary_node_mask_no_actives():
    neighbors = sgrid.neighbor_node_array(
        (2, 3), inactive=-2,
        closed_boundary_nodes=np.arange(6))
    assert_array_equal(neighbors, - 2 * np.ones((6, 4)).T)


def test_boundary_node_mask_3x3():
    neighbors = sgrid.neighbor_node_array(
        (3, 3), inactive=-2,
        closed_boundary_nodes=[0, 1, 2, 3, 5, 6, 7, 8])
    assert_array_equal(neighbors, -2 * np.ones((9, 4)).T)


def test_boundary_node_mask():
    X = -1
    neighbors = sgrid.neighbor_node_array((3, 3),
                                          inactive=X, closed_boundary_nodes=[0, 2, 3, 5, 6, 7, 8])
    expected = np.array([[X, X, X, X], [X, 4, X, X], [X, X, X, X],
                         [X, X, X, X], [X, X, X, 1], [X, X, X, X],
                         [X, X, X, X], [X, X, X, X], [X, X, X, X]])
    assert_array_equal(neighbors, expected.T)


def test_open_boundary():
    X = -1
    neighbors = sgrid.neighbor_node_array(
        (3, 3), closed_boundary_nodes=[3, 5, 6, 7, 8],
        open_boundary_nodes=[0, 1, 2], inactive=X)
    expected = np.array([[X, X, X, X], [X, 4, X, X], [X, X, X, X],
                         [X, X, X, X], [X, X, X, 1], [X, X, X, X],
                         [X, X, X, X], [X, X, X, X], [X, X, X, X]])
    assert_array_equal(neighbors.T, expected)


def test_open_boundary_4x5():
    X = -1
    neighbors = sgrid.neighbor_node_array(
        (4, 5), closed_boundary_nodes=[5, 10, 15, 16, 17, 18, 19],
        open_boundary_nodes=[0, 1, 2, 3, 4, 9, 14], inactive=X)
    expected = np.array([
        [X, X, X, X], [X, 6, X, X], [X, 7, X, X], [X, 8, X, X], [X, X, X, X],
        [X, X, X, X], [7, 11, X, 1], [8, 12, 6, 2], [9, 13, 7, 3], [X, X, 8, X],
        [X, X, X, X], [12, X, X, 6], [13, X, 11, 7], [
            14, X, 12, 8], [X, X, 13, X],
        [X, X, X, X], [X, X, X, X], [X, X, X, X], [X, X, X, X], [X, X, X, X],
    ])
    assert_array_equal(neighbors.T, expected)


def test_open_boundary_on_perimeter():
    X = -1

    neighbors = sgrid.neighbor_node_array((5, 4),
                                          open_boundary_nodes=sgrid.boundary_nodes(
                                              (5, 4)),
                                          inactive=X)

    expected = np.array([
        [X, X, X, X], [X, 5, X, X], [X, 6, X, X], [X, X, X, X],
        [5, X, X, X], [6, 9, 4, 1], [7, 10, 5, 2], [X, X, 6, X],
        [9, X, X, X], [10, 13, 8, 5], [11, 14, 9, 6], [X, X, 10, X],
        [13, X, X, X], [14, 17, 12, 9], [15, 18, 13, 10], [X, X, 14, X],
        [X, X, X, X], [X, X, X, 13], [X, X, X, 14], [X, X, X, X],
    ])
    assert_array_equal(neighbors, expected.T)


def test_closed_boundary_on_perimeter():
    X = -1

    neighbors = sgrid.neighbor_node_array((5, 4),
                                          closed_boundary_nodes=sgrid.boundary_nodes(
                                              (5, 4)),
                                          inactive=X)

    expected = np.array([
        [X, X, X, X], [X, X, X, X], [X, X, X, X], [X, X, X, X],
        [X, X, X, X], [6, 9, X, X], [X, 10, 5, X], [X, X, X, X],
        [X, X, X, X], [10, 13, X, 5], [X, 14, 9, 6], [X, X, X, X],
        [X, X, X, X], [14, X, X, 9], [X, X, 13, 10], [X, X, X, X],
        [X, X, X, X], [X, X, X, X], [X, X, X, X], [X, X, X, X],
    ])
    assert_array_equal(neighbors, expected.T)


def test_closed_boundaries_in_middle():
    X = -1

    neighbors = sgrid.neighbor_node_array((5, 4),
                                          closed_boundary_nodes=sgrid.boundary_nodes(
                                              (5, 4)),
                                          inactive=X)

    expected = np.array([
        [X, X, X, X], [X, X, X, X], [X, X, X, X], [X, X, X, X],
        [X, X, X, X], [6, 9, X, X], [X, 10, 5, X], [X, X, X, X],
        [X, X, X, X], [10, 13, X, 5], [X, 14, 9, 6], [X, X, X, X],
        [X, X, X, X], [14, X, X, 9], [X, X, 13, 10], [X, X, X, X],
        [X, X, X, X], [X, X, X, X], [X, X, X, X], [X, X, X, X],
    ])
    assert_array_equal(neighbors, expected.T)
